import abc
from pynever.tensor import Tensor
import pynever.networks as networks
from typing import List, Optional
import pynever.strategies.abstraction as abst
import pynever.nodes as nodes
import time
import numpy as np


class Property(abc.ABC):
    """
    An abstract class used to represent a generic property for a NeuralNetwork.
    """


class SMTLIBProperty(Property):
    """
    A concrete class used to represent a generic property for a NeuralNetwork expressed as a SMTLIB query.

    Attributes
    ----------
    smtlib_path : str
        Filepath for the SMTLIB file in which the property is defined.

    """

    def __init__(self, smtlib_path: str):
        self.smtlib_path = smtlib_path


class LocalRobustnessProperty(Property):
    """
    A concrete class used to represent a local robustness property for a NeuralNetwork.
    Formally the property check if the counterexample (i.e., the adversarial example) exists, therefore
    when the verification strategy check such property it should return True if the adversarial example exist and
    false otherwise.

    Attributes
    ----------
    data : Tensor
        Original data used to determine the local robustness.
    target : int
        If targeted is True then it is the desired target for the adversarial, otherwise it is the correct target of
        data.
    targeted : bool
        Flag which is True if the robustness property is targeted, False otherwise.
    norm : str
        Norm type used to determine the local robustness. At present the only acceptable value is Linf.
    epsilon : float
        Magnitude of the acceptable perturbation.
    bounds: list
        List of (lower_bound, upper_bound) for the data.

    """

    def __init__(self, data: Tensor, target: int, targeted: bool, norm: str, epsilon: float, bounds: list):

        self.data = data
        self.target = target
        self.targeted = targeted
        if norm != "Linf":
            raise NotImplementedError
        self.norm = norm
        self.epsilon = epsilon
        self.bounds = bounds


class NeVerProperty(Property):
    """
    A concrete class used to represent a NeVer property for a NeuralNetwork. We assume that the hyperplane
    out_coef_mat * y <= out_bias_mat represent the unsafe region (i.e., the negation of the desired property).
    At present the input set must be defined as in_coef_mat * x <= in_bias_mat

    Attributes
    ----------
    in_coef_mat: Tensor
        Matrix of the coefficients for the input constraints.
    in_bias_mat: Tensor
        Matrix of the biases for the input constraints.
    out_coef_mat: List[Tensor]
        Matrixes of the coefficients for the output constraints.
    out_bias_mat: List[Tensor]
        Matrixes of the biases for the output constraints.

    """

    def __init__(self, in_coef_mat: Tensor, in_bias_mat: Tensor, out_coef_mat: List[Tensor],
                 out_bias_mat: List[Tensor]):

        self.in_coef_mat = in_coef_mat
        self.in_bias_mat = in_bias_mat
        self.out_coef_mat = out_coef_mat
        self.out_bias_mat = out_bias_mat


class VerificationStrategy(abc.ABC):
    """
    An abstract class used to represent a Verification Strategy.

    Methods
    ----------
    verify(NeuralNetwork, Property)
        Verify that the neural network of interest satisfy the property given as argument
        using a verification strategy determined in the concrete children.

    """

    @abc.abstractmethod
    def verify(self, network: networks.NeuralNetwork, prop: Property) -> (bool, Optional[Tensor]):
        """
        Verify that the neural network of interest satisfy the property given as argument
        using a verification strategy determined in the concrete children.

        Parameters
        ----------
        network : NeuralNetwork
            The neural network to train.
        prop : Dataset
            The property which the neural network must satisfy.

        Returns
        ----------
        bool
            True is the neural network satisfy the property, False otherwise.

        """
        pass


class NeverVerification(VerificationStrategy):

    """
    Class used to represent the Never verification strategy.

    Attributes
    ----------
    log_filepath: str
        Filepath for saving the log files of the verification procedure.

    neuron_relevance : bool
        Flag to determine the heuristic used to decide how to refine the abstraction. If True the refinement
        decision is taken at the neuron level. Otherwise it is taken on the single sub stars. See the class
        abstraction.AbsReLUNode for details.

    iqr_thresholding : bool
        Flag used to determine if use the iqr thresholding heuristic in the refinement

    iqr_mult : float
        Multiplier for the iqr value.

    refinement_percentage: float
        Percentage of the neurons (if neuron_relevance = True) or the stars (if neuron_relevance = False) which
        will use the complete version of the abstraction algorithm (instead of the overapproximate version).

    refinement_level: int
        Refinement level for the sigmoid abstraction.

    Methods
    ----------
    verify(NeuralNetwork, Property)
        Verify that the neural network of interest satisfy the property given as argument.
    """

    def __init__(self, log_filepath: str, neuron_relevance: bool, iqr_thresholding: bool,
                 iqr_mult: float, refinement_percentage: float, refinement_level: int):

        assert 0 <= refinement_percentage <= 1, "refinement_percentage must be a value between 0 and 1"

        self.log_filepath = log_filepath
        self.neuron_relevance = neuron_relevance
        self.refinement_percentage = refinement_percentage
        self.iqr_thresholding = iqr_thresholding
        self.iqr_mult = iqr_mult
        self.refinement_level = refinement_level

    def verify(self, network: networks.NeuralNetwork, prop: Property) -> (bool, Optional[Tensor]):

        assert isinstance(network, networks.SequentialNetwork), "Only sequential networks are supported at present"
        abst_networks = abst.AbsSeqNetwork("Abstract Network")

        current_node = network.get_first_node()
        while current_node is not None:

            if isinstance(current_node, nodes.FullyConnectedNode):
                abst_networks.add_node(abst.AbsFullyConnectedNode("ABST_" + current_node.identifier, current_node))

            elif isinstance(current_node, nodes.ReLUNode):
                abst_networks.add_node(abst.AbsReLUNode("ABST_" + current_node.identifier, current_node,
                                                        self.neuron_relevance, self.iqr_thresholding, self.iqr_mult,
                                                        self.refinement_percentage))

            elif isinstance(current_node, nodes.SigmoidNode):
                abst_networks.add_node(abst.AbsSigmoidNode("ABST_" + current_node.identifier, current_node,
                                                           self.refinement_level))

            else:
                raise NotImplementedError

            current_node = network.get_next_node(current_node)

        with open(self.log_filepath, "w") as log_file:
            areas_log = self.log_filepath.replace(".txt", "_areas.txt")
            with open(areas_log, "w") as area_log_file:

                if isinstance(prop, NeVerProperty):

                    input_star = abst.Star(prop.in_coef_mat, prop.in_bias_mat)
                    input_starset = abst.StarSet({input_star})
                    current_node = abst_networks.get_first_node()
                    output_starset = input_starset
                    while current_node is not None:

                        time_start = time.perf_counter()
                        output_starset = current_node.forward(output_starset)
                        time_end = time.perf_counter()

                        print(f"Computing starset for layer {current_node.identifier}. Current starset has dimension "
                              f"{len(output_starset.stars)}. Time to compute: {time_end - time_start}s.")
                        log_file.write(f"Computing starset for layer {current_node.identifier}. "
                                       f"Current starset has dimension {len(output_starset.stars)}."
                                       f"Time to compute: {time_end - time_start}s.\n")

                        if isinstance(current_node, abst.AbsReLUNode):
                            for areas in current_node.overapprox_areas:
                                if isinstance(areas, np.ndarray):
                                    area_log_file.write(f"{areas.tolist()}\n")
                                else:
                                    area_log_file.write(f"{areas}\n")

                        current_node = abst_networks.get_next_node(current_node)

                    out_coef_mat = prop.out_coef_mat
                    out_bias_mat = prop.out_bias_mat

                else:
                    raise NotImplementedError

                verified = True
                for i in range(len(out_coef_mat)):

                    for star in output_starset.stars:
                        out_coef = out_coef_mat[i]
                        out_bias = out_bias_mat[i]
                        temp_star = abst.intersect_with_halfspace(star, out_coef, out_bias)
                        if not temp_star.check_if_empty():
                            verified = False
                            # print(f"Star {k}: Unsafe")

                log_file.write(f"Verification Result: {verified}.\n")

        return verified



import abc
import numpy as np
from pynever.tensor import Tensor
import torchvision as tv
import torch


class Dataset(abc.ABC):
    """
    An abstract class used to represent a Dataset.

    Methods
    ----------
    get_training_set()
        Return the part of the dataset used as the Training set as a tuple of Tensors (Data and Targets).
        It must be implemented in the concrete classes.
    get_test_set()
        Return the part of the dataset used as the Test set as a tuple of Tensors (Data and Targets).
        It must be implemented in the concrete classes.
    add_training_sample((Tensor, Tensor))
        Add a new sample to the training set.
        It must be implemented in the concrete classes.
    add_test_sample((Tensor, Tensor))
        Add a new sample to the test set.
        It must be implemented in the concrete classes.

    """

    @abc.abstractmethod
    def get_training_set(self) -> (Tensor, Tensor):
        raise NotImplementedError

    @abc.abstractmethod
    def get_test_set(self) -> (Tensor, Tensor):
        raise NotImplementedError

    @abc.abstractmethod
    def add_training_sample(self, sample: (Tensor, Tensor)):
        raise NotImplementedError

    @abc.abstractmethod
    def add_test_sample(self, sample: (Tensor, Tensor)):
        raise NotImplementedError


class MNISTDataset(Dataset):
    """
    A concrete class used to represent the MNIST Dataset.

    Attributes
    ----------
    training_set : (Tensor, Tensor)
        Tuple of Tensors containing the data and the target of the part of the dataset forming the training set.
    test_set : (Tensor, Tensor)
        Tuple of Tensors containing the data and the target of the part of the dataset forming the test set.

    Methods
    ----------
    get_training_set()
        Return the part of the dataset used as the Training set as a tuple of Tensors (Data and Targets).
    get_test_set()
        Return the part of the dataset used as the Test set as a tuple of Tensors (Data and Targets).
    add_training_sample((Tensor, Tensor))
        Add a new sample to the training set.
        It must be implemented in the concrete classes.
    add_test_sample((Tensor, Tensor))
        Add a new sample to the test set.
        It must be implemented in the concrete classes.
    """

    def __init__(self):

        datapath = 'data/'

        transform = tv.transforms.Compose([
            tv.transforms.ToTensor(),
            tv.transforms.Normalize((0.5,), (0.5,)),
        ])

        train_loader = torch.utils.data.DataLoader(
            tv.datasets.MNIST(datapath, train=True, download=True, transform=transform), batch_size=1,
            shuffle=True)

        test_loader = torch.utils.data.DataLoader(
            tv.datasets.MNIST(datapath, train=False, transform=transform), batch_size=1, shuffle=True)

        data_list = []
        target_list = []
        for data, target in train_loader:
            data_np = data[0][0].view(-1).numpy()
            target_np = target.numpy()
            data_list.append(data_np)
            target_list.append(target_np)

        self.training_set = (np.array(data_list), np.array(target_list).reshape(-1))

        data_list = []
        target_list = []
        for data, target in test_loader:
            data_np = data[0][0].view(-1).numpy()
            target_np = target.numpy()
            data_list.append(data_np)
            target_list.append(target_np)

        self.test_set = (np.array(data_list), np.array(target_list).reshape(-1))

    def get_training_set(self) -> (Tensor, Tensor):
        """
        Procedure which returns the Training Set of the MNIST dataset: such set contains 60000 grayscale images.

        Returns
        -------
        (Tensor, Tensor)
            The training set as a tuple of Tensor.
        """
        return self.training_set

    def get_test_set(self) -> (Tensor, Tensor):
        """
        Procedure which returns the Test Set of the MNIST dataset: such set contains 10000 grayscale images.

        Returns
        -------
        (Tensor, Tensor)
            The test set as a tuple of Tensor.
        """
        return self.test_set

    def add_training_sample(self, sample: (Tensor, Tensor)):
        """
        Procedure which adds a new sample to the Training Set of the MNIST dataset.

        Parameters
        ----------
        sample : (Tensor, Tensor)
            The sample we wish to add to the set. The first element is the data and the second is the target.

        """
        self.training_set = (np.append(self.training_set[0], sample[0], axis=0),
                             np.append(self.training_set[1], sample[1]))

    def add_test_sample(self, sample: (Tensor, Tensor)):
        """
        Procedure which adds a new sample to the Training Set of the MNIST dataset.

        Parameters
        ----------
        sample : (Tensor, Tensor)
            The sample we wish to add to the set. The first element is the data and the second is the target.

        """
        self.test_set = (np.append(self.test_set[0], sample[0], axis=0),
                         np.append(self.test_set[1], sample[1]))


class FMNISTDataset(Dataset):
    """
    A concrete class used to represent the FMNIST Dataset.

    Attributes
    ----------
    training_set : (Tensor, Tensor)
        Tuple of Tensors containing the data and the target of the part of the dataset forming the training set.
    test_set : (Tensor, Tensor)
        Tuple of Tensors containing the data and the target of the part of the dataset forming the test set.

    Methods
    ----------
    get_training_set()
        Return the part of the dataset used as the Training set as a tuple of Tensors (Data and Targets).
    get_test_set()
        Return the part of the dataset used as the Test set as a tuple of Tensors (Data and Targets).
    add_training_sample((Tensor, Tensor))
        Add a new sample to the training set.
        It must be implemented in the concrete classes.
    add_test_sample((Tensor, Tensor))
        Add a new sample to the test set.
        It must be implemented in the concrete classes.
    """

    def __init__(self):

        datapath = 'data/'

        transform = tv.transforms.Compose([
            tv.transforms.ToTensor(),
            tv.transforms.Normalize((0.5,), (0.5,)),
        ])

        train_loader = torch.utils.data.DataLoader(
            tv.datasets.FashionMNIST(datapath, train=True, download=True, transform=transform), batch_size=1,
            shuffle=True)

        test_loader = torch.utils.data.DataLoader(
            tv.datasets.FashionMNIST(datapath, train=False, transform=transform), batch_size=1, shuffle=True)

        data_list = []
        target_list = []
        for data, target in train_loader:
            data_np = data[0][0].view(-1).numpy()
            target_np = target.numpy()
            data_list.append(data_np)
            target_list.append(target_np)

        self.training_set = (np.array(data_list), np.array(target_list).reshape(-1))

        data_list = []
        target_list = []
        for data, target in test_loader:
            data_np = data[0][0].view(-1).numpy()
            target_np = target.numpy()
            data_list.append(data_np)
            target_list.append(target_np)

        self.test_set = (np.array(data_list), np.array(target_list).reshape(-1))

    def get_training_set(self) -> (Tensor, Tensor):
        """
        Procedure which returns the Training Set of the MNIST dataset: such set contains 60000 grayscale images.

        Returns
        -------
        (Tensor, Tensor)
            The training set as a tuple of Tensor.
        """
        return self.training_set

    def get_test_set(self) -> (Tensor, Tensor):
        """
        Procedure which returns the Test Set of the MNIST dataset: such set contains 10000 grayscale images.

        Returns
        -------
        (Tensor, Tensor)
            The test set as a tuple of Tensor.
        """
        return self.test_set

    def add_training_sample(self, sample: (Tensor, Tensor)):
        """
        Procedure which adds a new sample to the Training Set of the MNIST dataset.

        Parameters
        ----------
        sample : (Tensor, Tensor)
            The sample we wish to add to the set. The first element is the data and the second is the target.

        """
        self.training_set = (np.append(self.training_set[0], sample[0], axis=0),
                             np.append(self.training_set[1], sample[1]))

    def add_test_sample(self, sample: (Tensor, Tensor)):
        """
        Procedure which adds a new sample to the Training Set of the MNIST dataset.

        Parameters
        ----------
        sample : (Tensor, Tensor)
            The sample we wish to add to the set. The first element is the data and the second is the target.

        """
        self.test_set = (np.append(self.test_set[0], sample[0], axis=0),
                         np.append(self.test_set[1], sample[1]))


class DynamicsJamesPos(Dataset):
    """
    A concrete class used to represent the Dynamic James Dataset presented in the paper
    "Challenging SMT solvers to verify neural networks" by Pulina and Tacchella (2012).

    Attributes
    ----------
    training_set : (Tensor, Tensor)
        Tuple of Tensors containing the data and the target of the part of the dataset forming the training set.
    test_set : (Tensor, Tensor)
        Tuple of Tensors containing the data and the target of the part of the dataset forming the test set.

    Methods
    ----------
    get_training_set()
        Return the part of the dataset used as the Training set as a tuple of Tensors (Data and Targets).
    get_test_set()
        Return the part of the dataset used as the Test set as a tuple of Tensors (Data and Targets).
    add_training_sample((Tensor, Tensor))
        Add a new sample to the training set.
        It must be implemented in the concrete classes.
    add_test_sample((Tensor, Tensor))
        Add a new sample to the test set.
        It must be implemented in the concrete classes.
    """

    def __init__(self, training_set_path: str, test_set_path: str):

        train_set = np.loadtxt(training_set_path, delimiter=",")
        t_set = np.loadtxt(test_set_path, delimiter=",")
        training_set = (train_set[:, 0:8], train_set[:, 8:])
        test_set = (t_set[:, 0:8], t_set[:, 8:])
        self.training_set = training_set
        self.test_set = test_set

    def get_training_set(self) -> (Tensor, Tensor):
        """
        Procedure which returns the Training Set of the MNIST dataset: such set contains 60000 grayscale images.

        Returns
        -------
        (Tensor, Tensor)
            The training set as a tuple of Tensor.
        """
        return self.training_set

    def get_test_set(self) -> (Tensor, Tensor):
        """
        Procedure which returns the Test Set of the MNIST dataset: such set contains 10000 grayscale images.

        Returns
        -------
        (Tensor, Tensor)
            The test set as a tuple of Tensor.
        """
        return self.test_set

    def add_training_sample(self, sample: (Tensor, Tensor)):
        """
        Procedure which adds a new sample to the Training Set of the MNIST dataset.

        Parameters
        ----------
        sample : (Tensor, Tensor)
            The sample we wish to add to the set. The first element is the data and the second is the target.

        """
        self.training_set = (np.append(self.training_set[0], sample[0], axis=0),
                             np.append(self.training_set[1], sample[1]))

    def add_test_sample(self, sample: (Tensor, Tensor)):
        """
        Procedure which adds a new sample to the Training Set of the MNIST dataset.

        Parameters
        ----------
        sample : (Tensor, Tensor)
            The sample we wish to add to the set. The first element is the data and the second is the target.

        """
        self.test_set = (np.append(self.test_set[0], sample[0], axis=0),
                         np.append(self.test_set[1], sample[1]))



import abc
import pynever.networks as networks
import pynever.datasets as datasets
import pynever.strategies.conversion as cv
import pynever.strategies.training as training
import torch
import math
import torch.nn as nn


class PruningStrategy(abc.ABC):
    """
    An abstract class used to represent a Pruning Strategy.

    Methods
    ----------
    prune(NeuralNetwork, Dataset)
        Prune the neural network of interest using a pruning strategy determined in the concrete children.

    """

    @abc.abstractmethod
    def prune(self, network: networks.NeuralNetwork, dataset: datasets.Dataset) -> networks.NeuralNetwork:
        """
        Prune the neural network of interest using a pruning strategy determined in the concrete children.

        Parameters
        ----------
        network : NeuralNetwork
            The neural network to prune.
        dataset: Dataset
            The dataset to use for the pre-training and fine-tuning procedure.

        Returns
        ----------
        NeuralNetwork
            The Neural Network resulting from the application of the pruning strategy to the original network.

        """
        pass


class WeightPruning(PruningStrategy):
    """
    A concrete class used to represent the weight pruning strategy.
    This kind of pruning select the least important weights of the neural network
    of interest and set them to 0.
    We refer to https://arxiv.org/abs/1506.02626 for theoretical details on the strategy.

    Attributes
    ----------
    sparsity_rate : float
        It determines the percentage of neurons which will be removed. It must be a Real number between 0 and 1.
    training_strategy : AdamTraining
        The training strategy to use for pre-training and/or fine-tuning.
    pre_training : bool
        Flag to indicate if the network need to be pre-trained.

    Methods
    ----------
    prune(NeuralNetwork, Dataset)
        Prune the neural network of interest using the pruning strategy Weight Pruning.

    """

    def __init__(self, sparsity_rate: float, training_strategy: training.AdamTraining = None,
                 pre_training: bool = False):

        self.sparsity_rate = sparsity_rate
        self.training_strategy = training_strategy
        self.pre_training = pre_training

    def prune(self, network: networks.NeuralNetwork, dataset: datasets.Dataset) -> networks.NeuralNetwork:
        """
        Prune the neural network of interest using the pruning strategy Weight Pruning.

        Parameters
        ----------
        network : NeuralNetwork
            The neural network to prune.
        dataset : Dataset
            The dataset to use for the pre-training and fine-tuning procedure.

        Returns
        ----------
        NeuralNetwork
            The Neural Network resulting from the application of the pruning strategy to the original network.

        """

        if self.training_strategy is not None and self.pre_training:
            fine_tuning = self.training_strategy.fine_tuning
            self.training_strategy.fine_tuning = False
            network = self.training_strategy.train(network, dataset)
            self.training_strategy.fine_tuning = fine_tuning

        pytorch_converter = cv.PyTorchConverter()
        py_net = pytorch_converter.from_neural_network(network)

        py_net = self.__pruning(py_net)

        network.alt_rep_cache.clear()
        network.alt_rep_cache.append(py_net)
        network.up_to_date = False

        if self.training_strategy is not None and self.training_strategy.fine_tuning:
            old_l1_decay = self.training_strategy.l1_decay
            old_batchnorm_decay = self.training_strategy.batchnorm_decay
            self.training_strategy.l1_decay = 0
            self.training_strategy.batchnorm_decay = 0
            network = self.training_strategy.train(network, dataset)
            self.training_strategy.l1_decay = old_l1_decay
            self.training_strategy.batchnorm_decay = old_batchnorm_decay

        return network

    def __pruning(self, net: cv.PyTorchNetwork):
        """
        Procedure for the pruning of the weights of the PyTorchNetwork passed as an argument.

        Parameters
        ----------
        net : PyTorchNetwork
            The PyTorchNetwork to prune.

        Returns
        ----------
        PyTorchNetwork
            The pruned PyTorchNetwork.

        """

        # We transfer the internal pytorch model to the CPU for the pruning procedure.
        net.pytorch_network.cpu()

        # We compute the number of weights in the network
        num_weights = 0
        for m in net.pytorch_network.modules():

            if isinstance(m, torch.nn.Linear):
                num_weights += m.weight.numel()

        # We copy all the absolute values of the weights in a new tensor and we sort in ascending order
        weights = torch.zeros(num_weights)
        index = 0
        for m in net.pytorch_network.modules():

            if isinstance(m, torch.nn.Linear):
                size = m.weight.numel()
                weights[index:(index + size)] = m.weight.view(-1).abs().clone()
                index += size

        ordered_weights, ordered_indexes = torch.sort(weights)

        # We determine the number of weights we need to set to 0, given the sparsity rate.
        threshold_index = math.floor(num_weights * self.sparsity_rate)

        # We select the weight absolute value we will use as threshold value given the threshold index.
        threshold_value = ordered_weights[threshold_index]

        # We set all the weights of the different layers to 0 if they are less or equal than the threshold value
        # (in absolute value)

        for m in net.pytorch_network.modules():

            if isinstance(m, torch.nn.Linear):
                # The values of the mask are 0 when the corresponding weight is less then the threshold_value (in
                # absolute value), otherwise they are 1
                mask = m.weight.abs().gt(threshold_value).float()
                m.weight.data = torch.mul(m.weight, mask)

        return net


class NetworkSlimming(PruningStrategy):
    """
    A concrete class used to represent the network slimming pruning strategy.
    This kind of pruning select the least important neurons of the neural network
    of interest and eliminates them. It needs a batch normalization layer following each layer
    which should be pruned. We assume that the activation function is always applied after the batch
    normalization layer.
    We refer to https://arxiv.org/abs/1708.06519 for theoretical details on the strategy.

    Attributes
    ----------
    sparsity_rate : float
        It determines the percentage of neurons which will be removed. It must be a Real number between 0 and 1.
    training_strategy : AdamTraining
        The training strategy to use for pre-training and/or fine-tuning.
    pre_training : bool
        Flag to indicate if the network need to be pre-trained.

    Methods
    ----------
    prune(NeuralNetwork, Dataset)
        Prune the neural network of interest using the pruning strategy Network Slimming.

    """

    def __init__(self, sparsity_rate: float, training_strategy: training.AdamTraining = None,
                 pre_training: bool = False):

        self.sparsity_rate = sparsity_rate
        self.training_strategy = training_strategy
        self.pre_training = pre_training

    def prune(self, network: networks.NeuralNetwork, dataset: datasets.Dataset) -> networks.NeuralNetwork:
        """
        Prune the neural network of interest using the pruning strategy Network Slimming.

        Parameters
        ----------
        network : NeuralNetwork
            The neural network to prune.
        dataset: Dataset
            The dataset to use for the pre-training and fine-tuning procedure.

        Returns
        ----------
        NeuralNetwork
            The Neural Network resulting from the application of the pruning strategy to the original network.

        """

        if self.training_strategy is not None and self.pre_training:
            fine_tuning = self.training_strategy.fine_tuning
            self.training_strategy.fine_tuning = False
            network = self.training_strategy.train(network, dataset)
            self.training_strategy.fine_tuning = fine_tuning

        pytorch_converter = cv.PyTorchConverter()
        py_net = pytorch_converter.from_neural_network(network)

        py_net = self.__pruning(py_net)

        network.alt_rep_cache.clear()
        network.alt_rep_cache.append(py_net)
        network.up_to_date = False

        if self.training_strategy is not None and self.training_strategy.fine_tuning:
            old_l1_decay = self.training_strategy.l1_decay
            old_batchnorm_decay = self.training_strategy.batchnorm_decay
            self.training_strategy.l1_decay = 0
            self.training_strategy.batchnorm_decay = 0
            network = self.training_strategy.train(network, dataset)
            self.training_strategy.l1_decay = old_l1_decay
            self.training_strategy.batchnorm_decay = old_batchnorm_decay

        return network

    def __pruning(self, net: cv.PyTorchNetwork):
        """
        Procedure for the pruning of the neurons of the PyTorchNetwork passed as an argument.

        Parameters
        ----------
        net : PyTorchNetwork
            The PyTorchNetwork to prune.

        Returns
        ----------
        PyTorchNetwork
            The PyTorchNetwork resulting from the application of the pure pruning procedure.

        """

        # We transfer the internal pytorch model to the CPU for the pruning procedure.
        net.pytorch_network.cpu()

        # We compute the total number of weights in the batch normalization layers (which, for fully connected networks,
        # is equal to the number of neurons in the corresponding fully-connected layer).
        num_bn_weights = 0
        for m in net.pytorch_network.modules():
            if isinstance(m, torch.nn.BatchNorm1d):
                num_bn_weights += m.weight.numel()

        # We copy all the absolute values of the batch norm weights in a new tensor and we sort in ascending order
        bn_weights = torch.zeros(num_bn_weights)
        bn_weights_index = 0
        for m in net.pytorch_network.modules():
            if isinstance(m, torch.nn.BatchNorm1d):
                size = m.weight.numel()
                bn_weights[bn_weights_index:(bn_weights_index + size)] = m.weight.abs().clone()
                bn_weights_index += size

        ordered_bn_weights, ordered_bn_indexes = torch.sort(bn_weights)

        # We determine the number of neurons we need to remove, given the sparsity rate.
        threshold_index = math.floor(num_bn_weights * self.sparsity_rate)

        # We select the batch norm weight absolute value we will use as threshold value given the threshold index.
        threshold_value = ordered_bn_weights[threshold_index]

        # We now need to create a new network with the correct number of neurons in the different layers.
        # To do so we assume that in the network after a linear layer there is always a batch norm layer.

        new_layers = []
        previous_layer_mask = None
        old_layers = [m for m in net.pytorch_network.modules()]
        num_layers = len(old_layers)
        for i in range(num_layers):

            if (i == num_layers - 1) and isinstance(old_layers[i], torch.nn.Linear):

                # In this case we are considering the last layer of the network (which we assume to be a linear layer),
                # therefore the number of output of the new layer will be equal to the one of the old layer.

                previous_nonzero_indexes = previous_layer_mask.nonzero(as_tuple=True)[0]

                # If the old linear layer had bias then also the new linear layer has them.
                if old_layers[i].bias is None:
                    has_bias = False
                else:
                    has_bias = True

                # The number of input features for the new linear layer is equal to the number of non-zero elements in
                # the mask of the previous layer.
                num_in_features = int(previous_layer_mask.sum().item())

                # We create the new linear layer with the correct architecture.
                print(num_in_features, old_layers[i].out_features, has_bias)
                new_linear_layer = torch.nn.Linear(num_in_features, old_layers[i].out_features, has_bias)

                # We copy the parameters corresponding to the still existing neurons.
                new_linear_layer.weight.data = old_layers[i].weight[:, previous_nonzero_indexes].clone()

                if has_bias:
                    new_linear_layer.bias.data = old_layers[i].bias.data

                # We save the new linear layer.
                new_layers.append(new_linear_layer)

            elif isinstance(old_layers[i], torch.nn.Linear) and isinstance(old_layers[i + 1], torch.nn.BatchNorm1d):

                # If the layer old_layers[i] is the first linear layer then the previous layer mask corrspond to the
                # complete input.
                if previous_layer_mask is None:
                    previous_layer_mask = torch.ones(old_layers[i].in_features)

                # We compute the mask corresponding to the batch normalization layer.
                layer_mask = old_layers[i + 1].weight.abs().gt(threshold_value).float()
                new_neuron_number = int(layer_mask.sum().item())

                # We compute the indexes of the non-zero weights for the current batch norm layer and the previous one.
                current_nonzero_indexes = layer_mask.nonzero(as_tuple=True)[0]
                previous_nonzero_indexes = previous_layer_mask.nonzero(as_tuple=True)[0]

                # We create the new batch norm layer with the new neuron number.

                new_bn_layer = torch.nn.BatchNorm1d(new_neuron_number, eps=old_layers[i + 1].eps,
                                                    momentum=old_layers[i + 1].momentum,
                                                    affine=old_layers[i + 1].affine,
                                                    track_running_stats=old_layers[i + 1].track_running_stats)

                # We copy the parameters corresponding to the still existing neurons from the old batch norm layer
                # to the new one. They are identified by the indexes in current_nonzero_indexes.

                new_bn_layer.weight.data = old_layers[i + 1].weight[current_nonzero_indexes].clone()
                new_bn_layer.bias.data = old_layers[i + 1].bias[current_nonzero_indexes].clone()
                new_bn_layer.running_mean = old_layers[i + 1].running_mean[current_nonzero_indexes].clone()
                new_bn_layer.running_var = old_layers[i + 1].running_var[current_nonzero_indexes].clone()

                # If the old linear layer had bias then also the new linear layer has them.
                if old_layers[i].bias is None:
                    has_bias = False
                else:
                    has_bias = True

                # The number of input features for the new linear layer is equal to the number of non-zero elements in
                # the mask of the previous layer.
                num_in_features = int(previous_layer_mask.sum().item())

                # We create the new linear layer with the correct architecture.
                new_linear_layer = torch.nn.Linear(num_in_features, new_neuron_number, has_bias)

                # We copy the parameters corresponding to the still existing neurons.
                new_linear_layer.weight.data = old_layers[i].weight[current_nonzero_indexes, :].clone()
                new_linear_layer.weight.data = new_linear_layer.weight[:, previous_nonzero_indexes].clone()

                if has_bias:
                    new_linear_layer.bias.data = old_layers[i].bias[current_nonzero_indexes]

                # We save the new layers in the order in which they should be in our sequential model: first the linear
                # layer and then the batch norm layer.
                new_layers.append(new_linear_layer)
                new_layers.append(new_bn_layer)

                # We update the value of previous_layer_mask with the current mask.
                previous_layer_mask = layer_mask

            elif isinstance(old_layers[i], torch.nn.Linear) and not isinstance(old_layers[i + 1], torch.nn.BatchNorm1d):

                # If the layer old_layers[i] is the first linear layer then the previous layer mask correspond to the
                # complete input.
                if previous_layer_mask is None:
                    previous_layer_mask = torch.ones(old_layers[i].in_features)

                # If the linear layer is not followed by a batch normalization layer then it will not be neuron pruned,
                # therefore the layer_mask will be equals to the number of output features of the old layer.
                layer_mask = torch.ones(old_layers[i].out_features)

                # We compute the indexes of the non-zero weights for the current batch norm layer and the previous one.
                current_nonzero_indexes = layer_mask.nonzero(as_tuple=True)[0]
                previous_nonzero_indexes = previous_layer_mask.nonzero(as_tuple=True)[0]

                if old_layers[i].bias is None:
                    has_bias = False
                else:
                    has_bias = True

                # The number of input features for the new linear layer is equal to the number of non-zero elements in
                # the mask of the previous layer.
                num_in_features = previous_layer_mask.sum().item()

                # We create the new linear layer with the correct architecture.
                new_linear_layer = torch.nn.Linear(num_in_features, old_layers[i].out_features, has_bias)

                # We copy the parameters corresponding to the still existing neurons.
                new_linear_layer.weight.data = old_layers[i].weight[current_nonzero_indexes, :].clone()
                new_linear_layer.weight.data = new_linear_layer.weight[:, previous_nonzero_indexes].clone()

                if has_bias:
                    new_linear_layer.bias.data = old_layers[i].bias[current_nonzero_indexes]

                # We save the new layers in the order in which they should be in our sequential model: first the linear
                # layer and then the batch norm layer.
                new_layers.append(new_linear_layer)

            elif isinstance(old_layers[i], nn.ReLU):
                new_layers.append(nn.ReLU())

        pruned_network = torch.nn.Sequential(*new_layers)
        net.pytorch_network = pruned_network
        net.identifier = net.identifier + '_pruned'
        return net



import abc
import pynever.nodes as nodes


class NeuralNetwork(abc.ABC):
    """
    An abstract class used for our internal representation of a generic NeuralNetwork. It consists of a graph of LayerNodes
    and a list of AlternativeRepresentations. It should be noted that this data structure it is not a able
    to compute the input-output relation defined by the network. The properties of the computational graph are
    specialized in the concrete classes.

    Attributes
    ----------
    nodes : dict <str, LayerNode>
        Dictionary containing str keys and LayerNodes values. It contains the nodes of the graph,
        the identifier of the node of interest is used as a key in the nodes dictionary.
    edges : dict <str, list <str>>
        Dictionary of identifiers of LayerNodes, it contains for each nodes identified by the keys, the list of nodes
        connected to it.
    alt_rep_cache : dict <str, AlternativeRepresentation>
        Dictionary of containing str keys and AlternativeRepresentation values, it contains the
        AlternativeRepresentations of out network.
    up_to_date : bool
        Boolean Flag which shows if the internal representation is up-to-date with respect to the
        AlternativeRepresentations.

    """

    def __init__(self):

        self.nodes = {}
        self.edges = {}
        self.alt_rep_cache = []
        self.up_to_date = True


class SequentialNetwork(NeuralNetwork):
    """
    Concrete children of NeuralNetwork representing a sequential NeuralNetwork. It consists of a graph of LayerNodes
    and a list of AlternativeRepresentations. It should be noted that this data structure it is not a able
    to compute the input-output relation defined by the network. The computational graph of a SequentialNetwork must
    correspond to a standard list.

    Attributes
    ----------
    identifier : str
        Identifier of the Sequential Neural Network.

    Methods
    -------
    add_node(LayerNode)
        Procedure to add a new LayerNode to the sequential Neural Network.

    get_first_node()
        Procedure to extract the first node of the sequential Neural Network.

    get_next_node(LayerNode)
        Procedure to get the next node of the network given an input LayerNode

    get_last_node()
        Procedure to extract the last node of the sequential Neural Network.

    """

    def __init__(self, identifier: str):

        super().__init__()
        self.identifier = identifier

    def add_node(self, node: nodes.LayerNode):
        """
        Procedure to add a new LayerNode. In sequential network the new node must be connected directly to the
        previous node forming a list.

        Parameters
        ----------
        node : LayerNode
            New node to add to the Sequential network.

        """

        if len(self.nodes.keys()) == 0:
            self.nodes[node.identifier] = node
            self.edges[node.identifier] = []
        else:
            previous_node_key = self.get_last_node().identifier
            self.nodes[node.identifier] = node
            self.edges[previous_node_key].append(node.identifier)
            self.edges[node.identifier] = []

    def get_first_node(self) -> nodes.LayerNode:
        """
        Procedure to get the first LayerNode of the network.

        Return
        ---------
        LayerNode
            The first node of the network.

        """
        assert self.nodes

        keys = [key for key in self.nodes.keys()]
        for key in self.nodes.keys():
            for sub_key in self.nodes.keys():
                if sub_key in self.edges[key]:
                    keys.remove(sub_key)

        return self.nodes[keys[0]]

    def get_next_node(self, node: nodes.LayerNode) -> nodes.LayerNode:
        """
        Procedure to get the next LayerNode of the network given an input LayerNode.

        Return
        ---------
        LayerNode
            The next node of the network.

        """

        assert self.nodes

        next_node = None
        if node is not None:
            current_key = node.identifier
            if len(self.edges[current_key]) != 0:
                next_key = self.edges[current_key][0]
                next_node = self.nodes[next_key]
        else:
            next_node = self.get_first_node()

        return next_node

    def get_last_node(self) -> nodes.LayerNode:
        """
        Procedure to get the last LayerNode of the network.

        Return
        ---------
        LayerNode
            The last node of the network.

        """

        assert self.nodes

        current_node = self.get_first_node()
        while self.get_next_node(current_node) is not None:
            current_node = self.get_next_node(current_node)

        return current_node

    def __repr__(self):
        body = [node.to_string() for node in self.nodes.values()]
        return f"{self.identifier} : {body.__str__()}"

    def to_string(self):
        body = [node.to_string() for node in self.nodes.values()]
        return f"{self.identifier} : {body.__str__()}"



import pynever.strategies.conversion as cv
import pynever.nodes as nodes
import pynever.networks as networks
import pynever.datasets as datasets
import torch
import math
import torch.nn.functional as funct
from pynever.tensor import Tensor
import numpy as np


def combine_batchnorm1d(linear: nodes.FullyConnectedNode, batchnorm: nodes.BatchNorm1DNode) -> nodes.FullyConnectedNode:
    """
    Utility function to combine a BatchNorm1DNode node with a FullyConnectedNode in a corresponding FullyConnectedNode.

    Parameters
    ----------
    linear : FullyConnectedNode
        FullyConnectedNode to combine.
    batchnorm : BatchNorm1DNode
        BatchNorm1DNode to combine.

    Return
    ----------
    FullyConnectedNode
        The FullyConnectedNode resulting from the fusion of the two input nodes.

    """

    l_weight = torch.from_numpy(linear.weight)
    l_bias = torch.from_numpy(linear.bias)
    bn_running_mean = torch.from_numpy(batchnorm.running_mean)
    bn_running_var = torch.from_numpy(batchnorm.running_var)
    bn_weight = torch.from_numpy(batchnorm.weight)
    bn_bias = torch.from_numpy(batchnorm.bias)
    bn_eps = batchnorm.eps

    fused_bias = torch.div(bn_weight, torch.sqrt(bn_running_var + bn_eps))
    fused_bias = torch.mul(fused_bias, torch.sub(l_bias, bn_running_mean))
    fused_bias = torch.add(fused_bias, bn_bias)

    fused_weight = torch.diag(torch.div(bn_weight, torch.sqrt(bn_running_var + bn_eps)))
    fused_weight = torch.matmul(fused_weight, l_weight)

    fused_linear = nodes.FullyConnectedNode(linear.identifier, linear.in_features, linear.out_features, fused_weight.numpy(),
                                            fused_bias.numpy())

    return fused_linear


def combine_batchnorm1d_net(network: networks.SequentialNetwork) -> networks.SequentialNetwork:
    """
    Utilities function to combine all the FullyConnectedNodes followed by BatchNorm1DNodes in corresponding
    FullyConnectedNodes.

    Parameters
    ----------
    network : SequentialNetwork
        Sequential Network of interest of which we want to combine the nodes.

    Return
    ----------
    SequentialNetwork
        Corresponding Sequential Network with the combined nodes.

    """

    if not network.up_to_date:

        for alt_rep in network.alt_rep_cache:

            if alt_rep.up_to_date:

                if isinstance(alt_rep, cv.PyTorchNetwork):
                    pytorch_cv = cv.PyTorchConverter()
                    network = pytorch_cv.to_neural_network(alt_rep)
                elif isinstance(alt_rep, cv.ONNXNetwork):
                    onnx_cv = cv.ONNXConverter
                    network = onnx_cv.to_neural_network(alt_rep)
                else:
                    raise NotImplementedError
                break

    combined_network = networks.SequentialNetwork(network.identifier + '_combined')

    current_node = network.get_first_node()
    node_index = 1
    while network.get_next_node(current_node) is not None and current_node is not None:

        next_node = network.get_next_node(current_node)
        if isinstance(current_node, nodes.FullyConnectedNode) and isinstance(next_node, nodes.BatchNorm1DNode):
            combined_node = combine_batchnorm1d(current_node, next_node)
            combined_node.identifier = f"Combined_Linear_{node_index}"
            combined_network.add_node(combined_node)
            next_node = network.get_next_node(next_node)

        elif isinstance(current_node, nodes.FullyConnectedNode):
            identifier = f"Linear_{node_index}"
            new_node = nodes.FullyConnectedNode(identifier, current_node.in_features, current_node.out_features,
                                                current_node.weight, current_node.bias)
            combined_network.add_node(new_node)

        elif isinstance(current_node, nodes.ReLUNode):
            identifier = f"ReLU_{node_index}"
            new_node = nodes.ReLUNode(identifier, current_node.num_features)
            combined_network.add_node(new_node)
        else:
            raise NotImplementedError

        node_index += 1
        current_node = next_node

    if isinstance(current_node, nodes.FullyConnectedNode):
        identifier = f"Linear_{node_index}"
        new_node = nodes.FullyConnectedNode(identifier, current_node.in_features, current_node.out_features,
                                            current_node.weight, current_node.bias)
        combined_network.add_node(new_node)
    elif isinstance(current_node, nodes.ReLUNode):
        identifier = f"ReLU_{node_index}"
        new_node = nodes.ReLUNode(identifier, current_node.num_features)
        combined_network.add_node(new_node)
    else:
        raise NotImplementedError

    return combined_network


def testing(net: cv.PyTorchNetwork, dataset: datasets.Dataset, test_batch_size: int, cuda: bool) -> (float, float):
    """
    Testing procedure for a PyTorchNetwork.

    Parameters
    ----------
    net : PyTorchNetwork
        Neural Network to test.
    dataset : Dataset
        Dataset used for testing the network.
    test_batch_size : int
        Dimension for the test batch size for the testing procedure
    cuda : bool
        Whether to use the cuda library for the procedure (default: False).

    Returns
    ----------
    (float, float)
        Rate of correct samples and loss.

    """

    if cuda:
        net.pytorch_network.cuda()
    else:
        net.pytorch_network.cpu()

    test_set = dataset.get_test_set()

    net.pytorch_network.eval()
    net.pytorch_network.float()
    test_loss = 0
    correct = 0
    with torch.no_grad():

        batch_idx = 0
        data_idx = 0

        while data_idx < len(test_set[0]):

            if data_idx + test_batch_size >= len(test_set[0]):
                last_data_idx = len(test_set[0])
            else:
                last_data_idx = data_idx + test_batch_size

            data = torch.from_numpy(test_set[0][data_idx:last_data_idx, :])
            target = torch.from_numpy(test_set[1][data_idx:last_data_idx])

            if cuda:
                data, target = data.cuda(), target.cuda()

            data, target = torch.autograd.Variable(data), torch.autograd.Variable(target)
            output = net.pytorch_network(data)
            test_loss += funct.cross_entropy(output, target, reduction='sum').data.item()  # sum up batch loss
            pred = output.data.max(1, keepdim=True)[1]  # get the index of the max log-probability
            correct += pred.eq(target.data.view_as(pred)).cpu().sum().item()
            batch_idx += 1
            data_idx += test_batch_size

    test_loss /= float(math.floor(len(test_set[0]) / test_batch_size))
    print('\nTest set: Average loss: {:.4f}, Accuracy: {}/{} ({:.1f}%)\n'.format(
        test_loss, correct, len(test_set[0]),
        100. * correct / len(test_set[0])))

    return correct / float(len(test_set[0])), test_loss


def generate_targeted_linf_robustness_query(data: Tensor, adv_target: int, bounds: tuple,
                                            num_classes: int, epsilon: float, filepath: str):
    """
    Function to generate a targeted Robustness SMTLIB query and to save it to a SMTLIB file.
    The robustness query is of the kind based on the infinity norm.
    It assumes that the data and target are from a classification task.

    Parameters
    ----------
    data : Tensor
        Input data of interest.
    adv_target : int
        Desired adversarial target for the input data.
    bounds : (int, int)
        Bounds for the input data (lower_bound, upper_bound).
    num_classes : int
        Number of possible classes.
    epsilon : float
        Perturbation with respect to the infinity norm.
    filepath : str
        Filepath for the resulting SMTLIB file.

    """
    with open(filepath, "w") as f:

        flattened_data = data.flatten()
        for i in range(len(flattened_data)):
            f.write(f"(declare-const X_{i} Real)\n")

        for i in range(num_classes):
            f.write(f"(declare-const Y_{i} Real)\n")

        for i in range(len(flattened_data)):

            if flattened_data[i] - epsilon < bounds[0]:
                f.write(f"(assert (>= X_{i} {bounds[0]}))\n")
            else:
                f.write(f"(assert (>= X_{i} {flattened_data[i] - epsilon}))\n")

            if flattened_data[i] + epsilon > bounds[1]:
                f.write(f"(assert (<= X_{i} {bounds[1]}))\n")
            else:
                f.write(f"(assert (<= X_{i} {flattened_data[i] + epsilon}))\n")

        for i in range(num_classes):

            if i != adv_target:
                f.write(f"(assert (<= (- Y_{i} Y_{adv_target}) 0))\n")

def generate_untargeted_linf_robustness_query(data: Tensor, target: int, bounds: tuple,
                                              num_classes: int, epsilon: float, filepath: str):
    """
    Function to generate an untargeted Robustness SMTLIB query and to save it to a SMTLIB file.
    The robustness query is of the kind based on the infinity norm.
    It assumes that the data and target are from a classification task.

    Parameters
    ----------
    data : Tensor
        Input data of interest.
    adv_target : int
        Desired adversarial target for the input data.
    bounds : (int, int)
        Bounds for the input data (lower_bound, upper_bound).
    num_classes : int
        Number of possible classes.
    epsilon : float
        Perturbation with respect to the infinity norm.
    filepath : str
        Filepath for the resulting SMTLIB file.

    """
    with open(filepath, "w") as f:

        flattened_data = data.flatten()
        for i in range(len(flattened_data)):
            f.write(f"(declare-const X_{i} Real)\n")

        for i in range(num_classes):
            f.write(f"(declare-const Y_{i} Real)\n")

        for i in range(len(flattened_data)):

            if flattened_data[i] - epsilon < bounds[0]:
                f.write(f"(assert (>= X_{i} {bounds[0]}))\n")
            else:
                f.write(f"(assert (>= X_{i} {flattened_data[i] - epsilon}))\n")

            if flattened_data[i] + epsilon > bounds[1]:
                f.write(f"(assert (<= X_{i} {bounds[1]}))\n")
            else:
                f.write(f"(assert (<= X_{i} {flattened_data[i] + epsilon}))\n")

        output_query = "(assert (or"
        for i in range(num_classes):

            if i != target:
                output_query += f" (<= (- Y_{target} Y_{i}) 0)"

        output_query += "))"
        f.write(output_query)


def parse_linf_robustness_smtlib(filepath: str) -> (bool, list, int):
    """
    Function to extract the parameters of a robustness query from the smtlib file.
    It assume the SMTLIB file is structured as following:

        ; definition of the variables of interest
        (declare-const X_0 Real)
        (declare-const X_1 Real)
        ...
        (declare-const Y_1 Real)
        (declare-const Y_2 Real)
        ...
        ; definition of the constraints
        (assert (>= X_0 eps_0))
        (assert (<= X_0 eps_1))
        ...
        (assert (<= (- Y_0 Y_1) 0))
        ...

    Where the eps_i are Real numbers.

    Parameters
    ----------
    filepath : str
        Filepath to the SMTLIB file.

    Returns
    ----------
    (bool, list, int)
        Tuple of list: the first list contains the values eps_i for each variables as tuples (lower_bound, upper_bound),
        while the int correspond to the desired target for the related data.
    """
    targeted = True
    correct_target = -1
    lb = []
    ub = []
    with open(filepath, 'r') as f:

        for line in f:

            line = line.replace('(', '( ')
            line = line.replace(')', ' )')
            if line[0] == '(':
                aux = line.split()
                if aux[1] == 'assert':

                    if aux[4] == '(':
                        if aux[3] == 'or':
                            targeted = False
                            temp = aux[8].split("_")
                            correct_target = int(temp[1])
                        else:
                            targeted = True
                            temp = aux[7].split("_")
                            correct_target = int(temp[1])

                    else:

                        if aux[3] == ">=":
                            lb.append(float(aux[5]))
                        else:
                            ub.append(float(aux[5]))

    input_bounds = []
    for i in range(len(lb)):
        input_bounds.append((lb[i], ub[i]))

    return targeted, input_bounds, correct_target


def net_update(network: networks.NeuralNetwork) -> networks.NeuralNetwork:

    if not network.up_to_date:

        for alt_rep in network.alt_rep_cache:

            if alt_rep.up_to_date:
                if isinstance(alt_rep, cv.ONNXNetwork):
                    return cv.ONNXConverter().to_neural_network(alt_rep)
                elif isinstance(alt_rep, cv.PyTorchNetwork):
                    return cv.PyTorchConverter().to_neural_network(alt_rep)
                else:
                    raise NotImplementedError

    else:
        return network


def parse_acas_property(filepath: str) -> ((Tensor, Tensor), (Tensor, Tensor)):

    in_coeff = np.zeros((10, 5))
    in_bias = np.zeros((10, 1))
    out_coeff = []
    out_bias = []
    row_index = 0

    with open(filepath, 'r') as f:

        for line in f:

            if line[0] == "x":
                splitted_line = line.split(" ")
                var_index = int(splitted_line[0][1])
                if splitted_line[1] == ">=":
                    in_coeff[row_index, var_index] = -1
                    in_bias[row_index] = -float(splitted_line[2])
                else:
                    in_coeff[row_index, var_index] = 1
                    in_bias[row_index] = float(splitted_line[2])

            else:

                splitted_line = line.split(" ")
                if len(splitted_line) == 3:
                    var_index = int(splitted_line[0][1])
                    temp = np.zeros(5)
                    if splitted_line[1] == ">=":
                        temp[var_index] = -1
                        out_coeff.append(temp)
                        out_bias.append(-float(splitted_line[2]))
                    else:
                        temp[var_index] = 1
                        out_coeff.append(temp)
                        out_bias.append(float(splitted_line[2]))
                else:
                    var_index_1 = int(splitted_line[0][2])
                    var_index_2 = int(splitted_line[1][2])
                    temp = np.zeros(5)
                    if splitted_line[2] == ">=":
                        temp[var_index_1] = -1
                        temp[var_index_2] = 1
                        out_coeff.append(temp)
                        out_bias.append(-float(splitted_line[3]))
                    else:
                        temp[var_index_1] = 1
                        temp[var_index_2] = -1
                        out_coeff.append(temp)
                        out_bias.append(float(splitted_line[3]))

            row_index = row_index + 1

        out_coeff = np.array(out_coeff)
        array_out_bias = np.zeros((len(out_bias), 1))

        for i in range(len(out_bias)):
            array_out_bias[i, 0] = out_bias[i]

        out_bias = array_out_bias

    return (in_coeff, in_bias), (out_coeff, out_bias)


def parse_nnet(filepath: str) -> (list, list, list, list, list, list):

    with open(filepath) as f:

        line = f.readline()
        cnt = 1
        while line[0:2] == "//":
            line = f.readline()
            cnt += 1
        # numLayers does't include the input layer!
        numLayers, inputSize, outputSize, maxLayersize = [int(x) for x in line.strip().split(",")[:-1]]
        line = f.readline()

        # input layer size, layer1size, layer2size...
        layerSizes = [int(x) for x in line.strip().split(",")[:-1]]

        line = f.readline()
        symmetric = int(line.strip().split(",")[0])

        line = f.readline()
        inputMinimums = [float(x) for x in line.strip().split(",")[:-1]]

        line = f.readline()
        inputMaximums = [float(x) for x in line.strip().split(",")[:-1]]

        line = f.readline()
        means = [float(x) for x in line.strip().split(",")[:-1]]

        line = f.readline()
        ranges = [float(x) for x in line.strip().split(",")[:-1]]

        weights = []
        biases = []
        for layernum in range(numLayers):

            previousLayerSize = layerSizes[layernum]
            currentLayerSize = layerSizes[layernum + 1]
            # weights
            weights.append([])
            biases.append([])
            # weights
            for i in range(currentLayerSize):
                line = f.readline()
                aux = [float(x) for x in line.strip().split(",")[:-1]]
                weights[layernum].append([])
                for j in range(previousLayerSize):
                    weights[layernum][i].append(aux[j])
            # biases
            for i in range(currentLayerSize):
                line = f.readline()
                x = float(line.strip().split(",")[0])
                biases[layernum].append(x)

        numLayers = numLayers
        layerSizes = layerSizes
        inputSize = inputSize
        outputSize = outputSize
        maxLayersize = maxLayersize
        inputMinimums = inputMinimums
        inputMaximums = inputMaximums
        inputMeans = means[:-1]
        inputRanges = ranges[:-1]
        outputMean = means[-1]
        outputRange = ranges[-1]
        weights = weights
        biases = biases

        new_weights = []
        new_biases = []
        for i in range(numLayers):
            weight = np.array(weights[i])
            bias = np.array(biases[i])

            new_weights.append(weight)
            new_biases.append(bias)

        return new_weights, new_biases, inputMeans, inputRanges, outputMean, outputRange



import abc
import pynever.networks as networks
import pynever.datasets as datasets
import pynever.strategies.conversion as cv
import os
import shutil
import torch
import math
import numpy as np
import torch.nn as nn
import torch.nn.functional as funct
import pynever.utilities as utilities


class TrainingStrategy(abc.ABC):
    """
    An abstract class used to represent a Training Strategy.

    Methods
    ----------
    train(NeuralNetwork, Dataset)
        Train the neural network of interest using a training strategy determined in the concrete children.

    """

    @abc.abstractmethod
    def train(self, network: networks.NeuralNetwork, dataset: datasets.Dataset) -> networks.NeuralNetwork:
        """
        Train the neural network of interest using a pruning strategy determined in the concrete children.

        Parameters
        ----------
        network : NeuralNetwork
            The neural network to train.
        dataset : Dataset
            The dataset to use to train the neural network.

        Returns
        ----------
        NeuralNetwork
            The Neural Network resulting from the training of the original network using the training strategy and the
            dataset.

        """
        pass


class AdamTraining(TrainingStrategy):
    """
    A concrete class used to represent the Adam training strategy.
    This kind of training is based on an Adam optimizer.
    We refer to https://arxiv.org/abs/1412.6980 for theoretical details on the optimization algorithm.

    Attributes
    ----------
    n_epochs : int
        Number of epochs for the training procedure.
    train_batch_size : int
        Dimension for the train batch size for the training procedure
    test_batch_size : int
        Dimension for the test batch size for the training procedure
    learning_rate : float
        Learning rate parameter for the fine tuning procedure.
    betas : Tuple (float, float), optional
        Coefficients used for computing running averages of gradient and its square (default: (0.9, 0.999)).
    eps : float, optional
        Term added to the denominator to improve numerical stability (default: 1e-8).
    weight_decay : float, optional
        Coefficient of the L2 norm regularizer of the Adam optimizer (default: 0).
    cuda : bool, optional
        Whether to use the cuda library for the procedure (default: False).
    train_patience : int, optional
        The number of epochs in which the loss may not decrease before the
        training procedure is interrupted (default: 10).
    scheduler_patience : int, optional
        The number of epochs in which the loss may not decrease before the
        scheduler decrease the learning rate (default: 3).
    batchnorm_decay : float, optional
        It is the coefficient of the L1 norm regularizer applied only to the weights of the batch
        normalization layers. It is a preparatory parameter for pruning strategies which leverages
        the coefficients of the batch normalization layers. It should be selected considering the
        sparsity rate of the related pruning procedure (default: 0).
    l1_decay: float, optional
        Coefficient of the L1 norm regularizer. It should not be used with the weight_decay regularizer.
        It is also a preparatory parameter for pruning strategies which leverages the near-to-zero value of
        the weights (default: 0).
    fine_tuning : bool, optional
        Whether the training procedure should use the fine tuning routine (i.e., the weight with value = 0 are
        not updated by the optimizer (default: False).

    Methods
    ----------
    train(NeuralNetwork, Dataset)
        Train the neural network of interest using the training strategy Adam Training and the dataset passed as an
        argument.

    """

    def __init__(self, n_epochs: int, train_batch_size: int, test_batch_size: int, learning_rate: float,
                 betas: (float, float) = (0.9, 0.999), eps: float = 1e-8, weight_decay: float = 0, cuda: bool = False,
                 train_patience: int = 10, scheduler_patience: int = 3, batchnorm_decay: float = 0,
                 l1_decay: float = 0, fine_tuning: bool = False):

        self.n_epochs = n_epochs
        self.train_batch_size = train_batch_size
        self.test_batch_size = test_batch_size
        self.learning_rate = learning_rate
        self.betas = betas
        self.eps = eps
        self.weight_decay = weight_decay
        self.cuda = cuda
        self.train_patience = train_patience
        self.scheduler_patience = scheduler_patience
        self.batchnorm_decay = batchnorm_decay
        self.l1_decay = l1_decay
        self.fine_tuning = fine_tuning

    def train(self, network: networks.NeuralNetwork, dataset: datasets.Dataset) -> networks.NeuralNetwork:
        """
        Train the neural network of interest using the training strategy SGD Training.

        Parameters
        ----------
        network : NeuralNetwork
            The neural network to train.
        dataset : Dataset
            The dataset to use for the training of the network.

        Returns
        ----------
        NeuralNetwork
            The Neural Network resulting from the application of the training strategy to the original network.

        """

        pytorch_converter = cv.PyTorchConverter()
        py_net = pytorch_converter.from_neural_network(network)

        py_net = self.__training(py_net, dataset)

        network.alt_rep_cache.clear()
        network.alt_rep_cache.append(py_net)
        network.up_to_date = False

        return network

    def __training(self, net: cv.PyTorchNetwork, dataset: datasets.Dataset) -> cv.PyTorchNetwork:

        """
        Training procedure for the PyTorchNetwork.

        Parameters
        ----------
        net : PyTorchNetwork
            The PyTorchNetwork to train.
        dataset : Dataset
            The dataset to use for the training of the PyTorchNetwork

        Returns
        ----------
        PyTorchNetwork
            The trained PyTorchNetwork.

        """

        # If the training should be done with the GPU we set the model to cuda.
        if self.cuda:
            net.pytorch_network.cuda()
        else:
            net.pytorch_network.cpu()

        net.pytorch_network.train()
        net.pytorch_network.float()

        # We define the optimizer and the scheduler with the correct parameters.
        optimizer = torch.optim.Adam(params=net.pytorch_network.parameters(), lr=self.learning_rate, betas=self.betas,
                                     eps=self.eps, weight_decay=self.weight_decay)

        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, patience=self.scheduler_patience)

        start_epoch = 0
        training_set = dataset.get_training_set()

        # If a checkpoint exist, we load the checkpoint of interest
        checkpoints_path = 'checkpoints/' + net.identifier + '.pth.tar'
        best_model_path = 'checkpoints/' + net.identifier + '_best.pth.tar'
        if os.path.isfile(checkpoints_path):

            print("=> loading checkpoint '{}'".format(checkpoints_path))
            checkpoint = torch.load(checkpoints_path)
            start_epoch = checkpoint['epoch']
            best_prec1 = checkpoint['best_prec1']
            net.pytorch_network.load_state_dict(checkpoint['state_dict'])
            optimizer.load_state_dict(checkpoint['optimizer'])
            best_val_loss = checkpoint['best_val_loss']
            epochs_without_decrease = checkpoint['no_dec_epochs']
            print("=> loaded checkpoint '{}' (epoch {}) Prec1: {:f}"
                  .format(checkpoints_path, checkpoint['epoch'], best_prec1))

        else:
            print("=> no checkpoint found at '{}'".format(checkpoints_path))
            best_val_loss = 999
            epochs_without_decrease = 0

        history_score = np.zeros((self.n_epochs - start_epoch + 1, 3))

        # TRAINING

        best_prec1 = 0
        for epoch in range(start_epoch, self.n_epochs):

            if epochs_without_decrease > self.train_patience:
                break

            # EPOCH TRAINING

            net.pytorch_network.train()
            avg_loss = 0
            train_acc = 0
            batch_idx = 0
            data_idx = 0

            while data_idx < len(training_set[0]):

                if data_idx + self.train_batch_size >= len(training_set[0]):
                    last_data_idx = len(training_set[0])
                else:
                    last_data_idx = data_idx + self.train_batch_size

                data = torch.from_numpy(training_set[0][data_idx:last_data_idx, :])
                target = torch.from_numpy(training_set[1][data_idx:last_data_idx])

                if self.cuda:
                    data, target = data.cuda(), target.cuda()

                data, target = torch.autograd.Variable(data), torch.autograd.Variable(target)
                optimizer.zero_grad()
                output = net.pytorch_network(data)
                loss = funct.cross_entropy(output, target)
                avg_loss += loss.data.item()
                pred = output.data.max(1, keepdim=True)[1]
                train_acc += pred.eq(target.data.view_as(pred)).cpu().sum()
                loss.backward()

                # Pruning oriented training: it regularizes the batch norm coef values in order to identify unimportant
                # channels.
                for m in net.pytorch_network.modules():
                    if isinstance(m, nn.BatchNorm1d):
                        m.weight.grad.data.add_(self.batchnorm_decay * torch.sign(m.weight.data))

                # Pruning oriented training: it regularizes the weights in order to identify the unimportant ones.
                for m in net.pytorch_network.modules():

                    if isinstance(m, nn.Linear):
                        m.weight.grad.data.add_(self.l1_decay * torch.sign(m.weight.data))

                # If the fine_tuning flag is set then we assume that the training is used as fine tuning for a
                # weight pruning procedure, therefore the weight with value = 0 are not updated.
                if self.fine_tuning:

                    for m in net.pytorch_network.modules():

                        if isinstance(m, nn.Linear):
                            weight_copy = m.weight.data.abs().clone()
                            if self.cuda:
                                mask = weight_copy.gt(0).float().cuda()
                            else:
                                mask = weight_copy.gt(0).float()
                            m.weight.grad.data.mul_(mask)

                optimizer.step()
                if batch_idx % 100 == 0:
                    print('Train Epoch: {} [{}/{} ({:.1f}%)]\tLoss: {:.6f}'.format(
                        epoch, batch_idx * len(data), len(training_set[0]),
                               100. * batch_idx / math.floor(len(training_set[0]) / self.train_batch_size),
                        loss.data.item()))

                data_idx += self.train_batch_size
                batch_idx += 1

            history_score[epoch - start_epoch][0] = avg_loss / float(math.floor(len(training_set[0]) /
                                                                                self.train_batch_size))
            history_score[epoch - start_epoch][1] = train_acc / float(math.floor(len(training_set[0]) /
                                                                                 self.train_batch_size))

            # EPOCH TEST

            prec1, test_loss = utilities.testing(net, dataset, self.test_batch_size, self.cuda)

            if test_loss < best_val_loss:
                epochs_without_decrease = 0
                best_val_loss = test_loss
            else:
                epochs_without_decrease += 1

            if scheduler is not None:
                scheduler.step(test_loss)

            # CHECKPOINT

            history_score[epoch - start_epoch][2] = prec1
            is_best = prec1 > best_prec1
            best_prec1 = max(prec1, best_prec1)

            state = {
                'epoch': epoch + 1,
                'state_dict': net.pytorch_network.state_dict(),
                'best_prec1': best_prec1,
                'optimizer': optimizer.state_dict(),
                'best_val_loss': best_val_loss,
                'no_dec_epochs': epochs_without_decrease,
            }
            torch.save(state, checkpoints_path)
            if is_best:
                shutil.copyfile(checkpoints_path, best_model_path)

        print("Best accuracy: " + str(best_prec1))
        history_score[-1][0] = best_prec1

        if os.path.isfile(checkpoints_path):
            os.remove(checkpoints_path)

        if os.path.isfile(best_model_path):
            os.remove(best_model_path)

        return net


class AdamTrainingRegression(TrainingStrategy):
    """
    A concrete class used to represent the Adam training strategy for regression.
    This kind of training is based on an Adam optimizer.
    We refer to https://arxiv.org/abs/1412.6980 for theoretical details on the optimization algorithm.

    Attributes
    ----------
    n_epochs : int
        Number of epochs for the training procedure.
    train_batch_size : int
        Dimension for the train batch size for the training procedure
    test_batch_size : int
        Dimension for the test batch size for the training procedure
    learning_rate : float
        Learning rate parameter for the fine tuning procedure.
    betas : Tuple (float, float), optional
        Coefficients used for computing running averages of gradient and its square (default: (0.9, 0.999)).
    eps : float, optional
        Term added to the denominator to improve numerical stability (default: 1e-8).
    weight_decay : float, optional
        Coefficient of the L2 norm regularizer of the Adam optimizer (default: 0).
    cuda : bool, optional
        Whether to use the cuda library for the procedure (default: False).
    train_patience : int, optional
        The number of epochs in which the loss may not decrease before the
        training procedure is interrupted (default: 10).
    scheduler_patience : int, optional
        The number of epochs in which the loss may not decrease before the
        scheduler decrease the learning rate (default: 3).

    Methods
    ----------
    train(NeuralNetwork, Dataset)
        Train the neural network of interest using the training strategy Adam Training and the dataset passed as an
        argument.

    """

    def __init__(self, n_epochs: int, train_batch_size: int, test_batch_size: int, learning_rate: float,
                 betas: (float, float) = (0.9, 0.999), eps: float = 1e-8, weight_decay: float = 0, cuda: bool = False,
                 train_patience: int = 10, scheduler_patience: int = 3):

        self.n_epochs = n_epochs
        self.train_batch_size = train_batch_size
        self.test_batch_size = test_batch_size
        self.learning_rate = learning_rate
        self.betas = betas
        self.eps = eps
        self.weight_decay = weight_decay
        self.cuda = cuda
        self.train_patience = train_patience
        self.scheduler_patience = scheduler_patience

    def train(self, network: networks.NeuralNetwork, dataset: datasets.Dataset) -> networks.NeuralNetwork:
        """
        Train the neural network of interest using the training strategy SGD Training.

        Parameters
        ----------
        network : NeuralNetwork
            The neural network to train.
        dataset : Dataset
            The dataset to use for the training of the network.

        Returns
        ----------
        NeuralNetwork
            The Neural Network resulting from the application of the training strategy to the original network.

        """

        pytorch_converter = cv.PyTorchConverter()
        py_net = pytorch_converter.from_neural_network(network)

        py_net = self.__training(py_net, dataset)

        network.alt_rep_cache.clear()
        network.alt_rep_cache.append(py_net)
        network.up_to_date = False

        return network

    def __training(self, net: cv.PyTorchNetwork, dataset: datasets.Dataset) -> cv.PyTorchNetwork:

        """
        Training procedure for the PyTorchNetwork.

        Parameters
        ----------
        net : PyTorchNetwork
            The PyTorchNetwork to train.
        dataset : Dataset
            The dataset to use for the training of the PyTorchNetwork

        Returns
        ----------
        PyTorchNetwork
            The trained PyTorchNetwork.

        """

        # If the training should be done with the GPU we set the model to cuda.
        if self.cuda:
            net.pytorch_network.cuda()
        else:
            net.pytorch_network.cpu()

        net.pytorch_network.train()
        net.pytorch_network.double()

        # We define the optimizer and the scheduler with the correct parameters.
        optimizer = torch.optim.Adam(params=net.pytorch_network.parameters(), lr=self.learning_rate, betas=self.betas,
                                     eps=self.eps, weight_decay=self.weight_decay)

        scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, patience=self.scheduler_patience)

        start_epoch = 0
        training_set = dataset.get_training_set()

        # If a checkpoint exist, we load the checkpoint of interest
        checkpoints_path = 'checkpoints/' + net.identifier + '.pth.tar'
        best_model_path = 'checkpoints/' + net.identifier + '_best.pth.tar'
        if os.path.isfile(checkpoints_path):

            print("=> loading checkpoint '{}'".format(checkpoints_path))
            checkpoint = torch.load(checkpoints_path)
            start_epoch = checkpoint['epoch']
            best_prec1 = checkpoint['best_prec1']
            net.pytorch_network.load_state_dict(checkpoint['state_dict'])
            optimizer.load_state_dict(checkpoint['optimizer'])
            best_val_loss = checkpoint['best_val_loss']
            epochs_without_decrease = checkpoint['no_dec_epochs']
            print("=> loaded checkpoint '{}' (epoch {}) Prec1: {:f}"
                  .format(checkpoints_path, checkpoint['epoch'], best_prec1))

        else:
            print("=> no checkpoint found at '{}'".format(checkpoints_path))
            best_val_loss = 999
            epochs_without_decrease = 0

        history_score = np.zeros((self.n_epochs - start_epoch + 1, 3))

        # TRAINING

        best_prec1 = 999
        for epoch in range(start_epoch, self.n_epochs):

            if epochs_without_decrease > self.train_patience:
                break

            # EPOCH TRAINING

            net.pytorch_network.train()
            avg_loss = 0
            train_acc = 0
            batch_idx = 0
            data_idx = 0

            while data_idx < len(training_set[0]):

                if data_idx + self.train_batch_size >= len(training_set[0]):
                    last_data_idx = len(training_set[0])
                else:
                    last_data_idx = data_idx + self.train_batch_size

                data = torch.from_numpy(training_set[0][data_idx:last_data_idx, :])
                target = torch.from_numpy(training_set[1][data_idx:last_data_idx])

                if self.cuda:
                    data, target = data.cuda(), target.cuda()

                data, target = torch.autograd.Variable(data), torch.autograd.Variable(target)
                data = data.double()
                target = target.double()
                optimizer.zero_grad()
                output = net.pytorch_network(data)
                loss = funct.mse_loss(output, target)
                avg_loss += loss.data.item()
                loss.backward()

                optimizer.step()
                if batch_idx % 100 == 0:
                    print('Train Epoch: {} [{}/{} ({:.1f}%)]\tLoss: {:.6f}'.format(
                        epoch, batch_idx * len(data), len(training_set[0]),
                               100. * batch_idx / math.floor(len(training_set[0]) / self.train_batch_size),
                        loss.data.item()))

                data_idx += self.train_batch_size
                batch_idx += 1

            history_score[epoch - start_epoch][0] = avg_loss / float(math.floor(len(training_set[0]) /
                                                                                self.train_batch_size))
            history_score[epoch - start_epoch][1] = train_acc / float(math.floor(len(training_set[0]) /
                                                                                 self.train_batch_size))

            # EPOCH TEST

            test_set = dataset.get_test_set()

            net.pytorch_network.eval()
            net.pytorch_network.double()
            test_loss = 0
            with torch.no_grad():

                batch_idx = 0
                data_idx = 0

                while data_idx < len(test_set[0]):

                    if data_idx + self.test_batch_size >= len(test_set[0]):
                        last_data_idx = len(test_set[0])
                    else:
                        last_data_idx = data_idx + self.test_batch_size

                    data = torch.from_numpy(test_set[0][data_idx:last_data_idx, :])
                    target = torch.from_numpy(test_set[1][data_idx:last_data_idx])

                    if self.cuda:
                        data, target = data.cuda(), target.cuda()

                    data, target = torch.autograd.Variable(data), torch.autograd.Variable(target)
                    data = data.double()
                    target = target.double()
                    output = net.pytorch_network(data)
                    test_loss += funct.mse_loss(output, target).data.item()  # sum up batch loss
                    batch_idx += 1
                    data_idx += self.test_batch_size

            test_loss /= float(math.floor(len(test_set[0]) / self.test_batch_size))
            print('\nTest set: Average loss: {:.4f}\n'.format(test_loss))

            if test_loss < best_val_loss:
                epochs_without_decrease = 0
                best_val_loss = test_loss
            else:
                epochs_without_decrease += 1

            if scheduler is not None:
                scheduler.step(test_loss)

            # CHECKPOINT

            history_score[epoch - start_epoch][2] = test_loss
            is_best = test_loss < best_prec1
            best_prec1 = min(test_loss, best_prec1)

            state = {
                'epoch': epoch + 1,
                'state_dict': net.pytorch_network.state_dict(),
                'best_prec1': best_prec1,
                'optimizer': optimizer.state_dict(),
                'best_val_loss': best_val_loss,
                'no_dec_epochs': epochs_without_decrease,
            }

            torch.save(state, checkpoints_path)
            if is_best:
                shutil.copyfile(checkpoints_path, best_model_path)

        print("Best Loss: " + str(best_prec1))
        history_score[-1][0] = best_prec1

        if os.path.isfile(checkpoints_path):
            os.remove(checkpoints_path)

        if os.path.isfile(best_model_path):
            os.remove(best_model_path)

        return net



import copy
import abc
import pynever.networks as networks
import pynever.nodes as nodes
import torch
import onnx
import onnx.numpy_helper


class AlternativeRepresentation(abc.ABC):
    """
    An abstract class used to represent an alternative representation for a neural network.

    Attributes
    ----------
    identifier : str
        identifier for the alternative representation
    up_to_date : bool, optional
        flag which indicates if the alternative representation is up to date with respect
        to the internal representation of the network (optional: True).

    """
    def __init__(self, identifier: str, up_to_date: bool = True):
        self.identifier = identifier
        self.up_to_date = up_to_date


class ONNXNetwork(AlternativeRepresentation):
    """
    An class used to represent a ONNX representation for a neural network.

    Attributes
    ----------
    identifier : str
        identifier for the alternative representation
    onnx_network : onnx.ModelProto
        Real ONNX network.
    up_to_date : bool
        flag which indicates if the alternative representation is up to date with respect
        to the internal representation of the network (optional: True).

    """

    def __init__(self, identifier: str, onnx_network: onnx.ModelProto, up_to_date: bool = True):
        super().__init__(identifier, up_to_date)
        self.onnx_network = copy.deepcopy(onnx_network)


class PyTorchNetwork(AlternativeRepresentation):
    """
    An class used to represent a PyTorch representation for a neural network.

    Attributes
    ----------
    identifier : str
        identifier for the alternative representation
    pytorch_network : torch.nn.Module
        Real PyTorch network.
    up_to_date : bool
        flag which indicates if the alternative representation is up to date with respect
        to the internal representation of the network (optional: True).

    """
    def __init__(self, identifier: str, pytorch_network: torch.nn.Module, up_to_date: bool = True):
        super().__init__(identifier, up_to_date)
        self.pytorch_network = copy.deepcopy(pytorch_network)


class TensorflowNetwork(AlternativeRepresentation):
    """
    An class used to represent a Tensorflow representation for a neural network.

    Attributes
    ----------
    identifier : str
        identifier for the alternative representation
    up_to_date : bool
        flag which indicates if the alternative representation is up to date with respect
        to the internal representation of the network (optional: True).

    """
    def __init__(self, identifier: str, up_to_date: bool = True):
        super().__init__(identifier, up_to_date)


class ConversionStrategy(abc.ABC):
    """
    An abstract class used to represent a Conversion Strategy.

    Methods
    ----------
    from_neural_network(NeuralNetwork)
        Convert the neural network of interest to an alternative representation determined in the concrete children.
    to_neural_network(AlternativeRepresentation)
        Convert the alternative representation of interest to our internal representation of a Neural Network.

    """

    @abc.abstractmethod
    def from_neural_network(self, network: networks.NeuralNetwork) -> AlternativeRepresentation:
        """
        Convert the neural network of interest to an alternative representation determined in the concrete children.

        Parameters
        ----------
        network : NeuralNetwork
            The neural network to convert.

        Returns
        ----------
        AlternativeRepresentation
            The alternative representation resulting from the conversion of the original network.
        """
        pass

    @abc.abstractmethod
    def to_neural_network(self, alt_rep: AlternativeRepresentation) -> networks.NeuralNetwork:
        """
        Convert the alternative representation of interest to the internal one.

        Parameters
        ----------
        alt_rep : AlternativeRepresentation
            The Alternative Representation to convert.

        Returns
        ----------
        NeuralNetwork
            The Neural Network resulting from the conversion of Alternative Representation.
        """
        pass


class ONNXConverter(ConversionStrategy):
    """
    A class used to represent the conversion strategy for ONNX models.

    Methods
    ----------
    from_neural_network(NeuralNetwork)
        Convert the neural network of interest to a ONNXNetwork model.
    to_neural_network(ONNXNetwork)
        Convert the ONNXNetwork of interest to our internal representation of a Neural Network.

    """

    def from_neural_network(self, network: networks.NeuralNetwork) -> ONNXNetwork:
        """
        Convert the neural network of interest to a ONNX representation.

        Parameters
        ----------
        network : NeuralNetwork
            The neural network to convert.

        Returns
        ----------
        ONNXNetwork
            The ONNX representation resulting from the conversion of the original network.

        """

        alt_net = None
        for alt_rep in network.alt_rep_cache:
            if isinstance(alt_rep, ONNXNetwork) and alt_rep.up_to_date:
                alt_net = alt_rep

        if alt_net is None:

            if not network.up_to_date:

                for alt_rep in network.alt_rep_cache:

                    if alt_rep.up_to_date:

                        if isinstance(alt_rep, PyTorchNetwork):
                            pytorch_cv = PyTorchConverter()
                            network = pytorch_cv.to_neural_network(alt_rep)

                        else:
                            raise NotImplementedError
                        break

            if isinstance(network, networks.SequentialNetwork):

                current_node = None
                previous_output = 'X'
                input_info = []
                output_info = []
                initializers = []
                onnx_nodes = []

                while network.get_next_node(current_node) is not None:

                    current_node = network.get_next_node(current_node)
                    current_input = previous_output

                    if network.get_next_node(current_node) is None:
                        current_output = 'Y'
                    else:
                        current_output = current_node.identifier + '_output'

                    if isinstance(current_node, nodes.ReLUNode):

                        input_size = output_size = current_node.num_features
                        input_value_info = onnx.helper.make_tensor_value_info(current_input, onnx.TensorProto.FLOAT,
                                                                              [1, input_size])
                        output_value_info = onnx.helper.make_tensor_value_info(current_output, onnx.TensorProto.FLOAT,
                                                                               [1, output_size])

                        onnx_node = onnx.helper.make_node(
                            'Relu',
                            inputs=[current_input],
                            outputs=[current_output],
                        )

                        input_info.append(input_value_info)
                        output_info.append(output_value_info)
                        onnx_nodes.append(onnx_node)

                    elif isinstance(current_node, nodes.SigmoidNode):

                        input_size = output_size = current_node.num_features
                        input_value_info = onnx.helper.make_tensor_value_info(current_input, onnx.TensorProto.FLOAT,
                                                                              [1, input_size])
                        output_value_info = onnx.helper.make_tensor_value_info(current_output, onnx.TensorProto.FLOAT,
                                                                               [1, output_size])

                        onnx_node = onnx.helper.make_node(
                            'Sigmoid',
                            inputs=[current_input],
                            outputs=[current_output],
                        )

                        input_info.append(input_value_info)
                        output_info.append(output_value_info)
                        onnx_nodes.append(onnx_node)

                    elif isinstance(current_node, nodes.FullyConnectedNode):

                        input_size = current_node.in_features
                        output_size = current_node.out_features

                        input_weight = current_node.identifier + "_weight"
                        input_bias = current_node.identifier + "_bias"

                        input_value_info = onnx.helper.make_tensor_value_info(current_input, onnx.TensorProto.FLOAT,
                                                                              [1, input_size])
                        output_value_info = onnx.helper.make_tensor_value_info(current_output, onnx.TensorProto.FLOAT,
                                                                               [1, output_size])
                        weight_value_info = onnx.helper.make_tensor_value_info(input_weight, onnx.TensorProto.FLOAT,
                                                                               [output_size, input_size])
                        bias_value_info = onnx.helper.make_tensor_value_info(input_bias, onnx.TensorProto.FLOAT,
                                                                             [output_size])

                        # N.B: The Marabou procedure for reading ONNX models do not consider the attributes
                        # transA and transB, therefore we need to transpose the weight vector.
                        weight_tensor = onnx.numpy_helper.from_array(current_node.weight.T, input_weight)
                        bias_tensor = onnx.numpy_helper.from_array(current_node.bias, input_bias)

                        onnx_node = onnx.helper.make_node(
                            'Gemm',
                            inputs=[current_input, input_weight, input_bias],
                            outputs=[current_output],
                            alpha=1.0,
                            beta=1.0,
                            transA=0,
                            transB=0
                        )

                        input_info.append(input_value_info)
                        input_info.append(weight_value_info)
                        input_info.append(bias_value_info)

                        output_info.append(output_value_info)

                        initializers.append(weight_tensor)
                        initializers.append(bias_tensor)

                        onnx_nodes.append(onnx_node)

                    elif isinstance(current_node, nodes.BatchNorm1DNode):

                        input_size = output_size = current_node.num_features

                        input_scale = current_node.identifier + "_scale"
                        input_bias = current_node.identifier + "_bias"
                        input_mean = current_node.identifier + "_mean"
                        input_var = current_node.identifier + "_var"

                        input_value_info = onnx.helper.make_tensor_value_info(current_input, onnx.TensorProto.FLOAT,
                                                                              [1, input_size])
                        output_value_info = onnx.helper.make_tensor_value_info(current_output, onnx.TensorProto.FLOAT,
                                                                               [1, output_size])
                        scale_value_info = onnx.helper.make_tensor_value_info(input_scale, onnx.TensorProto.FLOAT,
                                                                              [input_size])
                        bias_value_info = onnx.helper.make_tensor_value_info(input_bias, onnx.TensorProto.FLOAT,
                                                                             [input_size])
                        mean_value_info = onnx.helper.make_tensor_value_info(input_mean, onnx.TensorProto.FLOAT,
                                                                             [input_size])
                        var_value_info = onnx.helper.make_tensor_value_info(input_var, onnx.TensorProto.FLOAT,
                                                                            [input_size])

                        scale_tensor = onnx.numpy_helper.from_array(current_node.weight, input_scale)
                        bias_tensor = onnx.numpy_helper.from_array(current_node.bias, input_bias)
                        mean_tensor = onnx.numpy_helper.from_array(current_node.running_mean, input_mean)
                        var_tensor = onnx.numpy_helper.from_array(current_node.running_var, input_var)

                        onnx_node = onnx.helper.make_node(
                            'BatchNormalization',
                            inputs=[current_input, input_scale, input_bias, input_mean, input_var],
                            outputs=[current_output],
                            epsilon=current_node.eps,
                            momentum=current_node.momentum
                        )

                        input_info.append(input_value_info)
                        input_info.append(scale_value_info)
                        input_info.append(bias_value_info)
                        input_info.append(mean_value_info)
                        input_info.append(var_value_info)

                        output_info.append(output_value_info)

                        initializers.append(scale_tensor)
                        initializers.append(bias_tensor)
                        initializers.append(mean_tensor)
                        initializers.append(var_tensor)

                        onnx_nodes.append(onnx_node)

                    else:
                        raise NotImplementedError

                    previous_output = current_output

                onnx_graph = onnx.helper.make_graph(
                    nodes=onnx_nodes,
                    name=network.identifier,
                    inputs=[input_info[0]],
                    outputs=[output_info[-1]],
                    initializer=initializers,
                    value_info=input_info
                )

                onnx_network = onnx.helper.make_model(graph=onnx_graph)
                alt_net = ONNXNetwork(network.identifier + "_onnx", onnx_network)

            else:
                raise NotImplementedError

        return alt_net

    def to_neural_network(self, alt_rep: ONNXNetwork) -> networks.NeuralNetwork:
        """
        Convert the ONNX representation of interest to the internal one.

        Parameters
        ----------
        alt_rep : ONNXNetwork
            The ONNX Representation to convert.

        Returns
        ----------
        NeuralNetwork
            The Neural Network resulting from the conversion of ONNX Representation.

        """
        identifier = alt_rep.identifier.replace("_onnx", "")
        network = networks.SequentialNetwork(identifier)

        parameters = {}
        for initializer in alt_rep.onnx_network.graph.initializer:
            parameters[initializer.name] = onnx.numpy_helper.to_array(initializer)

        shape_info = {}
        for value_info in alt_rep.onnx_network.graph.value_info:
            shape = []
            for dim in value_info.type.tensor_type.shape.dim:
                shape.append(dim.dim_value)
            shape_info[value_info.name] = shape

        node_index = 1
        for node in alt_rep.onnx_network.graph.node:

            if node.op_type == "Relu":

                # We assume that the real input of the node is always the first element of node.input
                # and that the shape is [batch_placeholder, real_size] for the inputs.
                num_features = shape_info[node.input[0]][1]
                network.add_node(nodes.ReLUNode(f"ReLU_{node_index}", num_features))

            elif node.op_type == "Sigmoid":
                num_features = shape_info[node.input[0]][1]
                network.add_node(nodes.SigmoidNode(f"Sigmoid_{node_index}", num_features))

            elif node.op_type == "Gemm":
                # We assume that the weight tensor is always the second element of node.input and the bias tensor
                # is always the third.
                # N.B: The Marabou procedure for reading ONNX models do not consider the attributes transA and transB,
                # therefore we need to transpose the weight vector.
                weight = parameters[node.input[1]].T
                bias = parameters[node.input[2]]
                in_features = weight.shape[1]
                out_features = weight.shape[0]
                network.add_node(nodes.FullyConnectedNode(f"Linear_{node_index}", in_features,
                                                          out_features, weight, bias))
            elif node.op_type == "BatchNormalization":
                # We assume that the real input is always the first element of node.input, the weight tensor
                # is always the second, the bias tensor is always the third, the running_mean always the fourth
                # and the running_var always the fifth.
                num_features = shape_info[node.input[0]][1]
                weight = parameters[node.input[1]]
                bias = parameters[node.input[2]]
                running_mean = parameters[node.input[3]]
                running_var = parameters[node.input[4]]
                # We assume that eps is always the first attribute and momentum is always the second.
                eps = node.attribute[0].f
                momentum = node.attribute[1].f
                network.add_node(nodes.BatchNorm1DNode(f"BatchNorm_{node_index}", num_features, weight, bias,
                                                       running_mean, running_var, eps, momentum))

            else:
                raise NotImplementedError

            node_index += 1

        return network


class PyTorchConverter(ConversionStrategy):
    """
    A class used to represent the conversion strategy for PyTorch models.

    Methods
    ----------
    from_neural_network(NeuralNetwork)
        Convert the neural network of interest to a PyTorchNetwork model.
    to_neural_network(PyTorchNetwork)
        Convert the PyTorchNetwork of interest to our internal representation of a Neural Network.

    """

    def from_neural_network(self, network: networks.NeuralNetwork) -> PyTorchNetwork:
        """
        Convert the neural network of interest to a PyTorch representation.

        Parameters
        ----------
        network : NeuralNetwork
            The neural network to convert.

        Returns
        ----------
        PyTorchNetwork
            The PyTorch representation resulting from the conversion of the original network.

        """

        alt_net = None
        pytorch_network = None
        for alt_rep in network.alt_rep_cache:
            if isinstance(alt_rep, PyTorchNetwork) and alt_rep.up_to_date:
                alt_net = alt_rep

        if alt_net is None:

            if not network.up_to_date:

                for alt_rep in network.alt_rep_cache:

                    if alt_rep.up_to_date:

                        if isinstance(alt_rep, ONNXNetwork):
                            onnx_cv = ONNXConverter()
                            network = onnx_cv.to_neural_network(alt_rep)

                        else:
                            raise NotImplementedError
                        break

            if isinstance(network, networks.SequentialNetwork):
                pytorch_layers = []
                for layer in network.nodes.values():

                    new_layer = None
                    if isinstance(layer, nodes.ReLUNode):
                        new_layer = torch.nn.ReLU()

                    elif isinstance(layer, nodes.SigmoidNode):
                        new_layer = torch.nn.Sigmoid()

                    elif isinstance(layer, nodes.FullyConnectedNode):

                        if layer.bias is not None:
                            has_bias = True
                        else:
                            has_bias = False

                        new_layer = torch.nn.Linear(in_features=layer.in_features,
                                                    out_features=layer.out_features,
                                                    bias=has_bias)

                        weight = torch.from_numpy(layer.weight)
                        new_layer.weight.data = weight

                        if has_bias:
                            bias = torch.from_numpy(layer.bias)
                            new_layer.bias.data = bias

                    elif isinstance(layer, nodes.BatchNorm1DNode):

                        new_layer = torch.nn.BatchNorm1d(num_features=layer.num_features,
                                                         eps=layer.eps, momentum=layer.momentum,
                                                         affine=layer.affine,
                                                         track_running_stats=layer.track_running_stats)

                        new_layer.weight.data = torch.from_numpy(layer.weight)
                        new_layer.bias.data = torch.from_numpy(layer.bias)
                        new_layer.running_mean.data = torch.from_numpy(layer.running_mean)
                        new_layer.running_var.data = torch.from_numpy(layer.running_var)

                    if new_layer is not None:
                        pytorch_layers.append(new_layer)

                pytorch_network = torch.nn.Sequential(*pytorch_layers)

            if alt_net is None and pytorch_network is None:
                print("WARNING: network to convert is not valid, the alternative representation is None")

            identifier = network.identifier + '_pytorch'
            alt_net = PyTorchNetwork(identifier=identifier, pytorch_network=pytorch_network)

        return alt_net

    def to_neural_network(self, alt_rep: PyTorchNetwork) -> networks.NeuralNetwork:
        """
        Convert the PyTorch representation of interest to the internal one.

        Parameters
        ----------
        alt_rep : PyTorchNetwork
            The PyTorch Representation to convert.

        Returns
        ----------
        NeuralNetwork
            The Neural Network resulting from the conversion of PyTorch Representation.

        """

        identifier = alt_rep.identifier.replace('_pytorch', '')
        network = networks.SequentialNetwork(identifier=identifier)

        node_index = 0
        size_prev_output = 0
        alt_rep.pytorch_network.cpu()
        for m in alt_rep.pytorch_network.modules():

            new_node = None

            if isinstance(m, torch.nn.ReLU):
                new_node = nodes.ReLUNode(identifier='ReLU_{}'.format(node_index), num_features=size_prev_output)

            elif isinstance(m, torch.nn.Sigmoid):
                new_node = nodes.SigmoidNode(identifier='Sigmoid_{}'.format(node_index), num_features=size_prev_output)

            elif isinstance(m, torch.nn.Linear):
                in_features = m.in_features
                out_features = m.out_features
                weight = m.weight.detach().numpy()
                bias = m.bias.detach().numpy()
                new_node = nodes.FullyConnectedNode(identifier='FullyConnected_{}'.format(node_index),
                                                    in_features=in_features, out_features=out_features,
                                                    weight=weight, bias=bias)

                size_prev_output = out_features

            elif isinstance(m, torch.nn.BatchNorm1d):

                num_features = m.num_features
                eps = m.eps
                momentum = m.momentum
                track_running_stats = m.track_running_stats
                affine = m.affine

                weight = m.weight.detach().numpy()
                bias = m.bias.detach().numpy()
                running_mean = m.running_mean.numpy()
                running_var = m.running_var.numpy()

                new_node = nodes.BatchNorm1DNode(identifier='BatchNorm1D_{}'.format(node_index),
                                                 num_features=num_features, weight=weight, bias=bias,
                                                 running_mean=running_mean, running_var=running_var, eps=eps,
                                                 momentum=momentum, affine=affine,
                                                 track_running_stats=track_running_stats)

                size_prev_output = num_features

            if new_node is not None:
                node_index += 1
                network.add_node(new_node)

        return network


class TensorflowConverter(ConversionStrategy):
    """
    A class used to represent the conversion strategy for Tensorflow models.

    Methods
    ----------
    from_neural_network(NeuralNetwork)
        Convert the neural network of interest to a TensorflowNetwork model.
    to_neural_network(ONNXNetwork)
        Convert the TensorflowNetwork of interest to our internal representation of a Neural Network.

    """

    def from_neural_network(self, network: networks.NeuralNetwork) -> TensorflowNetwork:
        """
        Convert the neural network of interest to a Tensorflow representation.

        Parameters
        ----------
        network : NeuralNetwork
            The neural network to convert.

        Returns
        ----------
        TensorflowNetwork
            The Tensorflow representation resulting from the conversion of the original network.

        """

    def to_neural_network(self, alt_rep: TensorflowNetwork) -> networks.NeuralNetwork:
        """
        Convert the Tensorflow representation of interest to the internal one.

        Parameters
        ----------
        alt_rep : TensorflowNetwork
            The Tensorflow Representation to convert.

        Returns
        ----------
        NeuralNetwork
            The Neural Network resulting from the conversion of Tensorflow Representation.

        """



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyNeVer",
    version="0.0.1.dev1",
    author="Dario Guidotti",
    author_email="dario.guidotti@edu.unige.it",
    license='GNU General Public License with Commons Clause License Condition v1.0',
    description="Package for the training, pruning and verification of neural networks.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/darioguidotti/pyNeVer",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Development Status :: 3 - Alpha",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=['numpy', 'scipy', 'ortools'],
)




import abc
from pynever.tensor import Tensor
import numpy as np


class LayerNode(abc.ABC):
    """
    An abstract class used for our internal representation of a generic Layer of a Neural Network.
    Its concrete children correspond to real network layers.

    Attributes
    ----------
    identifier : str
        Identifier of the LayerNode.

    """

    def __init__(self, identifier: str):

        self.identifier = identifier


class ReLUNode(LayerNode):
    """
    A class used for our internal representation of a ReLU Layer of a Neural Network.

    Attributes
    ----------
    identifier : str
        Identifier of the LayerNode.
    num_features : int
        Number of input and output features of the ReLU layer.
    """

    def __init__(self, identifier: str, num_features):
        super().__init__(identifier)
        self.num_features = num_features

    def __repr__(self):

        return f"ReLUNode - num_features = {self.num_features}"

    def to_string(self):

        return f"ReLUNode - num_features = {self.num_features}"


class SigmoidNode(LayerNode):
    """
    A class used for our internal representation of a Sigmoid Layer of a Neural Network.
    TO DO: It is not supported in the existing conversion strategies.

    Attributes
    ----------
    identifier : str
        Identifier of the LayerNode.
    num_features : int
        Number of input and output features of the Sigmoid layer.

    """

    def __init__(self, identifier: str, num_features):
        super().__init__(identifier)
        self.num_features = num_features

    def __repr__(self):

        return f"SigmoidNode - num_features = {self.num_features}"

    def to_string(self):

        return f"SigmoidNode - num_features = {self.num_features}"


class FullyConnectedNode(LayerNode):
    """
    A class used for our internal representation of a Fully Connected layer of a Neural Network

    Attributes
    ----------
    identifier : str
        Identifier of the LayerNode.
    in_features : int
        Number of input features of the fully connected layer.
    out_features : int
        Number of output features of the fully connected layer.
    weight : Tensor, optional
        Tensor containing the weight parameters of the fully connected layer.
    bias : Tensor, optional
        Tensor containing the bias parameters of the fully connected layer.
    """

    def __init__(self, identifier: str, in_features: int, out_features: int,
                 weight: Tensor = None, bias: Tensor = None):

        super().__init__(identifier)
        self.in_features = in_features
        self.out_features = out_features

        if weight is None:
            weight = np.random.normal(size=[out_features, in_features])

        if bias is None:
            bias = np.random.normal(size=[out_features])

        self.weight = weight
        self.bias = bias

    def __repr__(self):

        return f"FullyConnectedNode - in_features = {self.in_features}, out_features = {self.out_features}"

    def to_string(self):

        return f"FullyConnectedNode - in_features = {self.in_features}, out_features = {self.out_features}"


class BatchNorm1DNode(LayerNode):
    """
    A class used for our internal representation of a one dimensional Batch Normalization Layer.

    Attributes
    ----------
    identifier : str
        Identifier of the LayerNode.
    num_features : int
        Number of input and output feature of the Batch Normalization Layer.
    weight : Tensor, optional
        Tensor containing the weight parameters of the Batch Normalization Layer.
    bias : Tensor, optional
        Tensor containing the bias parameter of the Batch Normalization Layer.
    running_mean : Tensor, optional
        Tensor containing the running mean parameter of the Batch Normalization Layer.
    running_var : Tensor, optional
        Tensor containing the running var parameter of the Batch Normalization Layer.
    eps : float, optional
        Value added to the denominator for numerical stability (default: 1e-5).
    momentum : float, optional
        Value used for the running_mean and running_var computation. Can be set to None
        for cumulative moving average (default: 0.1)
    affine : bool, optional
        When set to True, the module has learnable affine parameter (default: True).
    track_running_stats : bool, optional
        When set to True, the module tracks the running mean and variance, when set to false the module
        does not track such statistics and always uses batch statistics in both training and eval modes (default: True).

    """

    def __init__(self, identifier: str, num_features: int, weight: Tensor = None, bias: Tensor = None,
                 running_mean: Tensor = None, running_var: Tensor = None, eps: float = 1e-5, momentum: float = 0.1,
                 affine: bool = True, track_running_stats: bool = True):

        super().__init__(identifier)
        self.num_features = num_features

        if track_running_stats and running_mean is None and running_var is None:
            running_mean = np.ones(num_features)
            running_var = np.zeros(num_features)

        if weight is None:
            weight = np.ones(num_features)

        if bias is None:
            bias = np.zeros(num_features)

        self.weight = weight
        self.bias = bias
        self.running_mean = running_mean
        self.running_var = running_var

        self.eps = eps
        self.momentum = momentum
        self.affine = affine
        self.track_running_stats = track_running_stats

    def __repr__(self):

        return f"BatchNorm1DNode - num_features = {self.num_features}"

    def to_string(self):
        return f"BatchNorm1DNode - num_features = {self.num_features}"



import abc
import pynever.nodes as nodes
import uuid
import numpy as np
import multiprocessing
import itertools
from typing import Set, List, Union
import math
import scipy.stats as sts

from pynever.tensor import Tensor
from ortools.linear_solver import pywraplp


# abstraction_logfile = "logs/ABSTRACTION_LOG.txt"


class AbsElement(abc.ABC):
    """
    An abstract class used for our internal representation of a generic Abstract Element (e.g., interval, zonotope,
    polyhedra etc.)

    Attributes
    ----------
    identifier : str
        Identifier of the AbsElement.
    """

    def __init__(self, identifier: str = None):
        if identifier is None:
            self.identifier = uuid.uuid4()
        else:
            self.identifier = identifier


class Star:
    """
    A concrete class used for our internal representation of a Star.
    The Star is defined as {x | x = c + Va such that Ca <= d}
    where c is a n-dimensional vector corresponding to the center of the Star.
    V is the n-by-m matrix composed by the basis vectors.
    a is the vector of m variables, C (p-by-m) and d (p-dim) are the matrix and the biases
    defining a set of constraints.

    We refer to <Star-Based Reachability Analysis of Deep Neural Networks>
    (https://link.springer.com/chapter/10.1007/978-3-030-30942-8_39) for details.

    Attributes
    ----------
    center : Tensor
        Center of the Star.
    basis_matrix : Tensor
        Matrix composed by the basis vectors of the Star
    predicate_matrix : Tensor
        Matrix of the Predicate.
    predicate_bias : Tensor
        Bias of the Predicate.
    ubs : list
        Upper bounds of the points defined by the Star.
    lbs : list
        Lower bounds of the points defined by the Star.
    is_empty : bool
        Boolean flag: True if the Star defines an empty set of points, False otherwise

    Methods
    ----------
    get_bounds()
        Function used to get the the upper and lower bounds of the n variables of the star.

    """

    def __init__(self, predicate_matrix: Tensor, predicate_bias: Tensor, center: Tensor = None,
                 basis_matrix: Tensor = None, lbs: list = None, ubs: list = None, is_empty: bool = None):

        predicate_dim_message = f"Error: the first dimension of the predicate_matrix ({predicate_matrix.shape[0]}) " \
                                f"must be equal to the dimension of the predicate_bias ({predicate_bias.shape[0]})."
        assert predicate_matrix.shape[0] == predicate_bias.shape[0], predicate_dim_message

        self.predicate_matrix = predicate_matrix
        self.predicate_bias = predicate_bias

        if center is None and basis_matrix is None:
            self.center = np.zeros((predicate_matrix.shape[1], 1))
            self.basis_matrix = np.identity(predicate_matrix.shape[1])

        else:
            center_dim_message = f"Error: the first dimension of the basis_matrix ({basis_matrix.shape[0]}) " \
                                 f"must be equal to the dimension of the center ({center.shape[0]})."
            assert center.shape[0] == basis_matrix.shape[0], center_dim_message

            basis_dim_message = f"Error: the second dimension of the basis_matrix ({basis_matrix.shape[1]}) " \
                                f"must be equal to the second dimension of the predicate_matrix " \
                                f"({predicate_matrix.shape[1]})."
            assert basis_matrix.shape[1] == predicate_matrix.shape[1], basis_dim_message

            self.center = center
            self.basis_matrix = basis_matrix

        if lbs is None:
            lbs = [None for i in range(self.center.shape[0])]

        if ubs is None:
            ubs = [None for i in range(self.center.shape[0])]

        self.lbs = lbs
        self.ubs = ubs
        self.is_empty = is_empty

    def check_if_empty(self) -> bool:
        """
        Function used to check if the set of points defined by the star is empty.

        Return
        ---------
        bool
            True if the star defines an empty set of points, False otherwise.
        """
        #start_time = time.perf_counter()
        if self.is_empty is None:

            solver, alphas, constraints = self.__get_predicate_lp_solver()
            objective = solver.Objective()
            for j in range(self.predicate_matrix.shape[1]):
                objective.SetCoefficient(alphas[j], 0)
            objective.SetOffset(0)

            objective.SetMinimization()
            status = solver.Solve()
            if status == pywraplp.Solver.INFEASIBLE or status == pywraplp.Solver.ABNORMAL:
                self.is_empty = True
            else:
                self.is_empty = False

        #end_time = time.perf_counter()
        #with open(abstraction_logfile, "a") as f:
        #    f.write(f"Empty,{end_time - start_time}\n")

        return self.is_empty

    def get_bounds(self, i) -> (float, float):
        """
        Function used to get the the upper and lower bounds of the n variables of the star.

        Return
        ---------
        (float, float)
            Tuple containing the lower and upper bounds of the variable i of the star
        """
        #start_time = time.perf_counter()
        if self.lbs[i] is None or self.ubs[i] is None:

            solver, alphas, constraints = self.__get_predicate_lp_solver()
            objective = solver.Objective()
            for j in range(self.basis_matrix.shape[1]):
                objective.SetCoefficient(alphas[j], self.basis_matrix[i, j])
            objective.SetOffset(self.center[i, 0])

            objective.SetMinimization()

            status = solver.Solve()

            """if status == pywraplp.Solver.ABNORMAL:
                objective = solver.Objective()
                for j in range(self.basis_matrix.shape[1]):
                    objective.SetCoefficient(alphas[j], 1)
                objective.SetOffset(0)
                objective.SetMinimization()
                status = solver.Solve()
                value = solver.Objective().Value()
                print(status)"""

            assert status == pywraplp.Solver.OPTIMAL, "The LP problem was not Optimal"

            lb = solver.Objective().Value()

            objective.SetMaximization()
            status = solver.Solve()
            #temp = solver.Objective().Value()
            assert status == pywraplp.Solver.OPTIMAL, "The LP problem was not Optimal"

            ub = solver.Objective().Value()

            self.lbs[i] = lb
            self.ubs[i] = ub

        else:
            lb = self.lbs[i]
            ub = self.ubs[i]

        #end_time = time.perf_counter()
        #with open(abstraction_logfile, "a") as f:
        #    f.write(f"Bound,{end_time - start_time}\n")
        return lb, ub

    def __get_predicate_lp_solver(self) -> (pywraplp.Solver, list, list):
        """
        Creates an lp solver using pywraplp and adds the variables and constraints
        corresponding to the predicate of the star.

        Returns
        ---------
        (pywraplp.Solver, list, list)
            Respectively the lp solver, the variables and the constraints.
        """

        solver = pywraplp.Solver.CreateSolver('GLOP')
        alphas = []
        for j in range(self.basis_matrix.shape[1]):
            new_alpha = solver.NumVar(-solver.infinity(), solver.infinity(), f'alpha_{j}')
            alphas.append(new_alpha)

        constraints = []
        for k in range(self.predicate_matrix.shape[0]):
            new_constraint = solver.Constraint(-solver.infinity(), self.predicate_bias[k, 0])
            for j in range(self.predicate_matrix.shape[1]):
                new_constraint.SetCoefficient(alphas[j], self.predicate_matrix[k, j])
            constraints.append(new_constraint)

        return solver, alphas, constraints


class StarSet(AbsElement):
    """
    Concrete class for our internal representation of a StarSet abstract element. A StarSet consist in a set
    of Star objects.

    Attributes
    ----------
    stars : Set[Star]
        Set of Star objects.
    """

    def __init__(self, stars: Set[Star] = None, identifier: str = None):
        super().__init__(identifier)
        if stars is None:
            self.stars = set()
        else:
            self.stars = stars


def intersect_with_halfspace(star: Star, coef_mat: Tensor, bias_mat: Tensor) -> Star:

    new_center = star.center
    new_basis_matrix = star.basis_matrix
    hs_pred_matrix = np.matmul(coef_mat, star.basis_matrix)
    hs_pred_bias = bias_mat - np.matmul(coef_mat, star.center)
    new_pred_matrix = np.vstack((star.predicate_matrix, hs_pred_matrix))
    new_pred_bias = np.vstack((star.predicate_bias, hs_pred_bias))

    new_star = Star(new_pred_matrix, new_pred_bias, new_center, new_basis_matrix)
    return new_star


def __mixed_step_relu(abs_input: Set[Star], var_index: int, refinement_flag: bool,
                      refinement_percentage: float = None) -> Set[Star]:

    abs_input = list(abs_input)
    abs_output = set()

    # Compute areas used to decide if the refinement is necessary.
    bounds = []
    t_areas = []
    for i in range(len(abs_input)):
        if abs_input[i].check_if_empty():
            bounds.append((None, None))
            t_areas.append(0)
        else:
            temp_lb, temp_ub = abs_input[i].get_bounds(var_index)
            bounds.append((temp_lb, temp_ub))
            if temp_lb < 0 and temp_ub > 0:
                t_areas.append(-temp_lb * temp_ub / 2.0)
            else:
                t_areas.append(0)

    t_areas = np.array(t_areas)

    # If the refinement_percentage is not None and the neuron should be refined I use the percentage
    # heuristic to select which stars I want to refine.
    if len(abs_input) == 1 and refinement_flag and refinement_percentage is not None and refinement_percentage > 0:
        ref_flags = [True]
    elif refinement_percentage is not None and refinement_flag:

        sorted_indexes = np.flip(np.argsort(t_areas))
        sorted_areas = t_areas[sorted_indexes]

        num_non_zero = 0
        for i in range(len(sorted_areas)):
            if sorted_areas[i] != 0:
                num_non_zero = num_non_zero + 1

        num_to_refine = int(refinement_percentage * num_non_zero)

        i = 0
        index_star_to_refine = []
        while i < num_to_refine:
            index_star_to_refine.append(sorted_indexes[i])
            i = i + 1

        ref_flags = []
        for i in range(len(t_areas)):
            if i in index_star_to_refine:
                ref_flags.append(True)
            else:
                ref_flags.append(False)

    # Otherwise I chose to refine or not refine them only based on the neuron refinement flag.
    else:
        ref_flags = [refinement_flag for i in range(len(abs_input))]

    for i in range(len(abs_input)):

        star = abs_input[i]
        ref_flag = ref_flags[i]

        if not star.check_if_empty():

            lb, ub = bounds[i]
            mask = np.identity(star.center.shape[0])
            mask[var_index, var_index] = 0

            if lb >= 0:
                abs_output = abs_output.union({star})

            elif ub <= 0:
                new_center = np.matmul(mask, star.center)
                new_basis_mat = np.matmul(mask, star.basis_matrix)
                new_pred_mat = star.predicate_matrix
                new_pred_bias = star.predicate_bias
                new_star = Star(new_pred_mat, new_pred_bias, new_center, new_basis_mat)
                abs_output = abs_output.union({new_star})

            else:

                if ref_flag:

                    # Creating lower bound star.
                    lower_star_center = np.matmul(mask, star.center)
                    lower_star_basis_mat = np.matmul(mask, star.basis_matrix)
                    # Adding x <= 0 constraints to the predicate.
                    lower_predicate_matrix = np.vstack((star.predicate_matrix, star.basis_matrix[var_index, :]))
                    # Possibile problema sulla dimensionalita' di star.center[var_index]
                    lower_predicate_bias = np.vstack((star.predicate_bias, -star.center[var_index]))
                    lower_star = Star(lower_predicate_matrix, lower_predicate_bias, lower_star_center,
                                      lower_star_basis_mat)

                    # Creating upper bound star.
                    upper_star_center = star.center
                    upper_star_basis_mat = star.basis_matrix
                    # Adding x >= 0 constraints to the predicate.
                    upper_predicate_matrix = np.vstack((star.predicate_matrix, -star.basis_matrix[var_index, :]))
                    # Possibile problema sulla dimensionalita' di star.center[var_index]
                    upper_predicate_bias = np.vstack((star.predicate_bias, star.center[var_index]))
                    upper_star = Star(upper_predicate_matrix, upper_predicate_bias, upper_star_center,
                                      upper_star_basis_mat)

                    abs_output = abs_output.union({lower_star, upper_star})

                else:

                    col_c_mat = star.predicate_matrix.shape[1]
                    row_c_mat = star.predicate_matrix.shape[0]

                    c_mat_1 = np.zeros((1, col_c_mat + 1))
                    c_mat_1[0, col_c_mat] = -1
                    c_mat_2 = np.hstack((np.array([star.basis_matrix[var_index, :]]), -np.ones((1, 1))))
                    coef_3 = - ub / (ub - lb)
                    c_mat_3 = np.hstack((np.array([coef_3 * star.basis_matrix[var_index, :]]), np.ones((1, 1))))
                    c_mat_0 = np.hstack((star.predicate_matrix, np.zeros((row_c_mat, 1))))

                    d_0 = star.predicate_bias
                    d_1 = np.zeros((1, 1))
                    d_2 = -star.center[var_index] * np.ones((1, 1))
                    d_3 = np.array([(ub / (ub - lb)) * (star.center[var_index] - lb)])

                    new_pred_mat = np.vstack((c_mat_0, c_mat_1, c_mat_2, c_mat_3))
                    new_pred_bias = np.vstack((d_0, d_1, d_2, d_3))

                    new_center = np.matmul(mask, star.center)
                    temp_basis_mat = np.matmul(mask, star.basis_matrix)
                    temp_vec = np.zeros((star.basis_matrix.shape[0], 1))
                    temp_vec[var_index, 0] = 1
                    new_basis_mat = np.hstack((temp_basis_mat, temp_vec))

                    new_star = Star(new_pred_mat, new_pred_bias, new_center, new_basis_mat)

                    abs_output = abs_output.union({new_star})

    return abs_output


def mixed_single_relu_forward(star: Star, neuron_relevance: bool, iqr_thresholding: bool, iqr_mult: float,
                              refinement_percentage: float) -> (Set[Star], Tensor):

    temp_abs_input = {star}
    if star.check_if_empty():
        return set()
    else:

        # If neuron_relevance is True it means that we want to chose which neurons will use the complete relu at this
        # level and not for the sub-stars generated during the computation of the various output neurons. Moreover it
        # means that the refinement_percentage is in relation with the number of neurons, not the number of sub-stars.
        n_areas = []
        if neuron_relevance:

            for i in range(star.center.shape[0]):
                lb, ub = star.get_bounds(i)
                if lb < 0 and ub > 0:
                    n_areas.append(-lb * ub / 2.0)
                else:
                    n_areas.append(0)

            n_areas = np.array(n_areas)

            sorted_indexes = np.flip(np.argsort(n_areas))
            sorted_areas = n_areas[sorted_indexes]

            num_non_zero = 0
            for i in range(len(sorted_areas)):
                if sorted_areas[i] > 0:
                    num_non_zero = num_non_zero + 1

            num_to_refine = int(refinement_percentage * num_non_zero)

            i = 0
            index_star_to_refine = []
            while i < num_to_refine:
                index_star_to_refine.append(sorted_indexes[i])
                i = i + 1

            refinement_flags = []
            for i in range(len(n_areas)):
                if i in index_star_to_refine:
                    refinement_flags.append(True)
                else:
                    refinement_flags.append(False)

            refinement_percentage = None

        elif iqr_thresholding:

            for i in range(star.center.shape[0]):
                lb, ub = star.get_bounds(i)
                if lb < 0 and ub > 0:
                    n_areas.append(-lb * ub / 2.0)
                else:
                    n_areas.append(0)

            n_areas = np.array(n_areas)

            threshold = np.median(n_areas) + iqr_mult * sts.iqr(n_areas)
            refinement_flags = []
            for i in range(len(n_areas)):
                if n_areas[i] > threshold:
                    refinement_flags.append(True)
                else:
                    refinement_flags.append(False)

            refinement_percentage = None

        # Otherwise the decision is left to the __mixed_step_relu function and the refinement_percentage is referred
        # to the number of sub-star. The refinement flags are set to True since potentially every neurons can be
        # refined.
        else:
            refinement_flags = [True for i in range(star.center.shape[0])]

        for i in range(star.center.shape[0]):
            temp_abs_input = __mixed_step_relu(temp_abs_input, i, refinement_flags[i], refinement_percentage)

        return temp_abs_input, n_areas


def single_fc_forward(star: Star, weight: Tensor, bias: Tensor) -> Set[Star]:
    """
    Utility function for the management of the forward for AbsFullyConnectedNode. It is outside
    the class scope since multiprocessing does not support parallelization with
    function internal to classes.
    """
    assert (weight.shape[1] == star.basis_matrix.shape[0])

    new_basis_matrix = np.matmul(weight, star.basis_matrix)
    new_center = np.matmul(weight, star.center) + bias
    new_predicate_matrix = star.predicate_matrix
    new_predicate_bias = star.predicate_bias

    new_star = Star(new_predicate_matrix, new_predicate_bias, new_center, new_basis_matrix)

    return {new_star}


def sig(x: float) -> float:
    return 1.0 / (1.0 + math.exp(-x))


def sig_fod(x: float) -> float:
    return math.exp(-x) / math.pow(1 + math.exp(-x), 2)


def area_sig_triangle(lb: float, ub: float) -> float:

    x_p = (ub * sig_fod(ub) - lb * sig_fod(lb)) / (sig_fod(ub) - sig_fod(lb)) - \
          (sig(ub) - sig(lb)) / (sig_fod(ub) - sig_fod(lb))

    y_p = sig_fod(ub) * (x_p - ub) + sig(ub)

    height = abs(y_p - (sig(ub) - sig(lb)) / (ub - lb) * x_p + sig(lb) - lb * (sig(ub) - sig(lb)) / (ub - lb)) / \
             math.sqrt(1 + math.pow((sig(ub) - sig(lb)) / (ub - lb), 2))

    base = math.sqrt(math.pow(ub - lb, 2) + math.pow(sig(ub) - sig(lb), 2))

    return base * height / 2.0


def __recursive_step_sigmoid(star: Star, var_index: int, approx_level: int, lb: float, ub: float, tolerance: float) -> Set[Star]:

    assert approx_level >= 0

    if abs(ub - lb) < tolerance:

        if ub <= 0:
            if ub + tolerance > 0:
                ub = 0
            else:
                ub = ub + tolerance
            lb = lb - tolerance
        else:
            if lb - tolerance < 0:
                lb = 0
            else:
                lb = lb - tolerance
            ub = ub + tolerance

    assert (lb <= 0 and ub <= 0) or (lb >= 0 and ub >= 0)

    mask = np.identity(star.center.shape[0])
    mask[var_index, var_index] = 0

    if approx_level == 0:

        if lb < 0 and ub <= 0:

            c_mat_1 = np.hstack((np.array([sig_fod(lb) * star.basis_matrix[var_index, :]]), -np.ones((1, 1))))
            c_mat_2 = np.hstack((np.array([sig_fod(ub) * star.basis_matrix[var_index, :]]), -np.ones((1, 1))))
            coef_3 = - (sig(ub) - sig(lb)) / (ub - lb)
            c_mat_3 = np.hstack((np.array([coef_3 * star.basis_matrix[var_index, :]]), np.ones((1, 1))))

            d_1 = np.array([-sig_fod(lb) * (star.center[var_index] - lb) - sig(lb)])
            d_2 = np.array([-sig_fod(ub) * (star.center[var_index] - ub) - sig(ub)])
            d_3 = np.array([-coef_3 * (star.center[var_index] - lb) + sig(lb)])

        else:

            c_mat_1 = np.hstack((np.array([-sig_fod(lb) * star.basis_matrix[var_index, :]]), np.ones((1, 1))))
            c_mat_2 = np.hstack((np.array([-sig_fod(ub) * star.basis_matrix[var_index, :]]), np.ones((1, 1))))
            coef_3 = (sig(ub) - sig(lb)) / (ub - lb)
            c_mat_3 = np.hstack((np.array([coef_3 * star.basis_matrix[var_index, :]]), -np.ones((1, 1))))

            d_1 = np.array([sig_fod(lb) * (star.center[var_index] - lb) + sig(lb)])
            d_2 = np.array([sig_fod(ub) * (star.center[var_index] - ub) + sig(ub)])
            d_3 = np.array([-coef_3 * (star.center[var_index] - lb) - sig(lb)])

        col_c_mat = star.predicate_matrix.shape[1]

        # Adding lb and ub bounds to enhance stability
        c_mat_lb = np.zeros((1, col_c_mat + 1))
        c_mat_lb[0, col_c_mat] = -1
        d_lb = -sig(lb) * np.ones((1, 1))

        c_mat_ub = np.zeros((1, col_c_mat + 1))
        c_mat_ub[0, col_c_mat] = 1
        d_ub = sig(ub) * np.ones((1, 1))

        row_c_mat = star.predicate_matrix.shape[0]
        c_mat_0 = np.hstack((star.predicate_matrix, np.zeros((row_c_mat, 1))))
        d_0 = star.predicate_bias

        new_pred_mat = np.vstack((c_mat_0, c_mat_1, c_mat_2, c_mat_3, c_mat_lb, c_mat_ub))
        new_pred_bias = np.vstack((d_0, d_1, d_2, d_3, d_lb, d_ub))

        new_center = np.matmul(mask, star.center)
        temp_basis_mat = np.matmul(mask, star.basis_matrix)
        temp_vec = np.zeros((star.basis_matrix.shape[0], 1))
        temp_vec[var_index, 0] = 1
        new_basis_mat = np.hstack((temp_basis_mat, temp_vec))

        new_star = Star(new_pred_mat, new_pred_bias, new_center, new_basis_mat)

        return {new_star}

    else:

        # We need to select the boundary between lb and ub. The optimal boundary is the one which minimizes the
        # area of the two resulting triangle. Since computing the optimal is too slow we do an approximate search
        # between lb and ub considering s search points.

        num_search_points = 10
        boundaries = np.linspace(lb, ub, num_search_points, endpoint=False)
        boundaries = boundaries[1:]

        best_boundary = None
        smallest_area = 99999999
        for boundary in boundaries:
            area_1 = area_sig_triangle(lb, boundary)
            area_2 = area_sig_triangle(boundary, ub)
            if area_1 + area_2 < smallest_area:
                smallest_area = area_1 + area_2
                best_boundary = boundary

        star_set = set()
        star_set = star_set.union(__recursive_step_sigmoid(star, var_index, approx_level - 1, lb, best_boundary, tolerance))
        star_set = star_set.union(__recursive_step_sigmoid(star, var_index, approx_level - 1, best_boundary, ub, tolerance))

        return star_set


def __approx_step_sigmoid(abs_input: Set[Star], var_index: int, approx_level: int, tolerance: float) -> Set[Star]:

    abs_output = set()
    for star in abs_input:

        if not star.check_if_empty():
            lb, ub = star.get_bounds(var_index)

            if (lb < 0) and (ub > 0):
                abs_output = abs_output.union(__recursive_step_sigmoid(star, var_index, approx_level, lb, 0,
                                                                       tolerance))
                abs_output = abs_output.union(__recursive_step_sigmoid(star, var_index, approx_level, 0, ub,
                                                                       tolerance))
            else:
                abs_output = abs_output.union(__recursive_step_sigmoid(star, var_index, approx_level, lb,
                                                                       ub, tolerance))

    return abs_output


def single_sigmoid_forward(star: Star, approx_levels: List[int]) -> Set[Star]:
    tolerance = 0.01
    temp_abs_input = {star}
    for i in range(star.center.shape[0]):
        temp_abs_input = __approx_step_sigmoid(temp_abs_input, i, approx_levels[i], tolerance)
        print(f"Index {i}, NumStar: {len(temp_abs_input)}")
    return temp_abs_input


class RefinementState(abc.ABC):
    """
    A class used for the internal control of the refinement strategies/heuristics applied in the abstraction refinement
    step.
    """
    pass


class AbsLayerNode(abc.ABC):
    """
    An abstract class used for our internal representation of a generic Abstract Transformer Layer of an
    AbsNeural Network. Its concrete children correspond to real abstract interpretation network layers.

    Attributes
    ----------
    identifier : str
        Identifier of the AbsLayerNode.

    ref_node : LayerNode
        LayerNode di riferimento per l'abstract transformer.

    Methods
    ----------
    forward(AbsElement)
        Function which takes an AbsElement and compute the corresponding output AbsElement based on the abstract
        transformer.

    backward(RefinementState)
        Function which takes a reference to the refinement state and update both it and the state of the abstract
        transformer to control the refinement component of the abstraction.

    """

    def __init__(self, identifier: str, ref_node: nodes.LayerNode):
        self.identifier = identifier
        self.ref_node = ref_node

    @abc.abstractmethod
    def forward(self, abs_input: AbsElement) -> AbsElement:
        """
        Compute the output AbsElement based on the input AbsElement and the characteristics of the
        concrete abstract transformer.

        Parameters
        ----------
        abs_input : AbsElement
            The input abstract element.

        Returns
        ----------
        AbsElement
            The AbsElement resulting from the computation corresponding to the abstract transformer.
        """
        pass

    @abc.abstractmethod
    def backward(self, ref_state: RefinementState):
        """
        Update the RefinementState.

        Parameters
        ----------
        ref_state: RefinementState
            The RefinementState to update.
        """
        pass


class AbsReLUNode(AbsLayerNode):
    """
    A class used for our internal representation of a ReLU Abstract transformer.

    Attributes
    ----------
    identifier : str
        Identifier of the LayerNode.

    ref_node : ReLUNode
        LayerNode di riferimento per l'abstract transformer.

    neuron_relevance : bool, optional
        Flag used to determine if the refinement should be done with relation to the neurons or the stars
        (default: True).

    iqr_thresholding : bool, optional
        Flag used to determine if the refinement should be done using the interquartile range of the areas of the
        overapproximation as a threshold.

    iqr_mult : float
        Multiplier for the iqr value.

    refinement_percentage : float, optional
        Percentage of the neurons (or stars) we chose to refine (default: 0).

    overapprox_areas : Tensor
        Tensor containing the areas of the overapproximation of the star passed in the forward with respect to the
        different neurons of the layer. It must have dimension number of stars x number of neurons.

    Methods
    ----------
    forward(AbsElement)
        Function which takes an AbsElement and compute the corresponding output AbsElement based on the abstract
        transformer.

    backward(RefinementState)
        Function which takes a reference to the refinement state and update both it and the state of the abstract
        transformer to control the refinement component of the abstraction.
    """

    def __init__(self, identifier: str, ref_node: nodes.ReLUNode, neuron_relevance: bool = True,
                 iqr_thresholding: bool = True, iqr_mult: float = 1, refinement_percentage: float = 0):

        super().__init__(identifier, ref_node)
        assert 0 <= refinement_percentage <= 1, "refinement_percentage must be a value between 0 and 1"

        self.neuron_relevance = neuron_relevance
        self.refinement_percentage = refinement_percentage
        self.overapprox_areas = None
        self.iqr_thresholding = iqr_thresholding
        self.iqr_mult = iqr_mult

    def forward(self, abs_input: AbsElement) -> AbsElement:
        """
        Compute the output AbsElement based on the input AbsElement and the characteristics of the
        concrete abstract transformer.

        Parameters
        ----------
        abs_input : AbsElement
            The input abstract element.

        Returns
        ----------
        AbsElement
            The AbsElement resulting from the computation corresponding to the abstract transformer.
        """

        # parallel is used for debug: it permits to return to a sequential procedure
        parallel = True
        if isinstance(abs_input, StarSet):
            if parallel:
                return self.__parallel_starset_forward(abs_input)
            else:
                return self.__starset_forward(abs_input)
        else:
            raise NotImplementedError

    def backward(self, ref_state: RefinementState):
        """
        Update the RefinementState.

        Parameters
        ----------
        ref_state: RefinementState
            The RefinementState to update.
        """
        pass

    def __parallel_starset_forward(self, abs_input: StarSet) -> StarSet:

        #with open(abstraction_logfile, "a") as f:
        #    f.write("Node: " + self.identifier + "\n")

        my_pool = multiprocessing.Pool(multiprocessing.cpu_count())
        parallel_results = my_pool.starmap(mixed_single_relu_forward, zip(abs_input.stars,
                                                                          itertools.repeat(self.neuron_relevance),
                                                                          itertools.repeat(self.iqr_thresholding),
                                                                          itertools.repeat(self.iqr_mult),
                                                                          itertools.repeat(self.refinement_percentage)))
        my_pool.close()

        abs_output = StarSet()
        self.overapprox_areas = []
        for result in parallel_results:
            star_areas = result[1]
            self.overapprox_areas.append(star_areas)
            star_set = result[0]
            abs_output.stars = abs_output.stars.union(star_set)

        return abs_output

    def __starset_forward(self, abs_input: StarSet) -> StarSet:
        """
        Forward function specialized for the concrete AbsElement StarSet.
        """

        abs_output = StarSet()
        for star in abs_input.stars:

            results = mixed_single_relu_forward(star, self.neuron_relevance, self.iqr_thresholding,
                                                self.iqr_mult, self.refinement_percentage)

            abs_output.stars = abs_output.stars.union(results[0])

        return abs_output


class AbsFullyConnectedNode(AbsLayerNode):
    """
    A class used for our internal representation of a Fully Connected Abstract transformer.

    Attributes
    ----------
    identifier : str
        Identifier of the LayerNode.

    ref_node : FullyConnectedNode
        LayerNode di riferimento per l'abstract transformer.

    Methods
    ----------
    forward(AbsElement)
        Function which takes an AbsElement and compute the corresponding output AbsElement based on the abstract
        transformer.

    backward(RefinementState)
        Function which takes a reference to the refinement state and update both it and the state of the abstract
        transformer to control the refinement component of the abstraction.
    """

    def __init__(self, identifier: str, ref_node: nodes.FullyConnectedNode):
        super().__init__(identifier, ref_node)

    def forward(self, abs_input: AbsElement) -> AbsElement:
        """
        Compute the output AbsElement based on the input AbsElement and the characteristics of the
        concrete abstract transformer.

        Parameters
        ----------
        abs_input : AbsElement
            The input abstract element.

        Returns
        ----------
        AbsElement
            The AbsElement resulting from the computation corresponding to the abstract transformer.
        """
        if isinstance(abs_input, StarSet):
            return self.__starset_forward(abs_input)
        else:
            raise NotImplementedError

    def __starset_forward(self, abs_input: StarSet) -> StarSet:

        my_pool = multiprocessing.Pool(multiprocessing.cpu_count())
        # Need to expand bias since they are memorized like one-dimensional vector in fullyconnectednodes.
        if self.ref_node.bias.shape != (self.ref_node.weight.shape[0], 1):
            bias = np.expand_dims(self.ref_node.bias, 1)
        else:
            bias = self.ref_node.bias
        parallel_results = my_pool.starmap(single_fc_forward, zip(abs_input.stars,
                                                                  itertools.repeat(self.ref_node.weight),
                                                                  itertools.repeat(bias)))
        my_pool.close()
        abs_output = StarSet()
        for star_set in parallel_results:
            abs_output.stars = abs_output.stars.union(star_set)

        return abs_output

    def backward(self, ref_state: RefinementState):
        """
        Update the RefinementState.

        Parameters
        ----------
        ref_state: RefinementState
            The RefinementState to update.
        """
        pass


class AbsSigmoidNode(AbsLayerNode):
    """
    A class used for our internal representation of a Sigmoid transformer.

    Attributes
    ----------
    identifier : str
        Identifier of the LayerNode.

    ref_node : SigmoidNode
        LayerNode di riferimento per l'abstract transformer.

    refinement_level : Union[int, List[int]]

    Methods
    ----------
    forward(AbsElement)
        Function which takes an AbsElement and compute the corresponding output AbsElement based on the abstract
        transformer.

    backward(RefinementState)
        Function which takes a reference to the refinement state and update both it and the state of the abstract
        transformer to control the refinement component of the abstraction.
    """

    def __init__(self, identifier: str, ref_node: nodes.SigmoidNode, approx_levels: Union[int, List[int]] = None):
        super().__init__(identifier, ref_node)

        if approx_levels is None:
            approx_levels = [0 for i in range(ref_node.num_features)]
        elif isinstance(approx_levels, int):
            approx_levels = [approx_levels for i in range(ref_node.num_features)]

        self.approx_levels = approx_levels

    def forward(self, abs_input: AbsElement) -> AbsElement:
        """
        Compute the output AbsElement based on the input AbsElement and the characteristics of the
        concrete abstract transformer.

        Parameters
        ----------
        abs_input : AbsElement
            The input abstract element.

        Returns
        ----------
        AbsElement
            The AbsElement resulting from the computation corresponding to the abstract transformer.
        """
        if isinstance(abs_input, StarSet):
            return self.__starset_forward(abs_input)
        else:
            raise NotImplementedError

    def __starset_forward(self, abs_input: StarSet) -> StarSet:

        parallel = True
        if parallel:
            abs_output = StarSet()
            my_pool = multiprocessing.Pool(1)
            parallel_results = my_pool.starmap(single_sigmoid_forward, zip(abs_input.stars,
                                                                           itertools.repeat(self.approx_levels)))
            my_pool.close()
            for star_set in parallel_results:
                abs_output.stars = abs_output.stars.union(star_set)
        else:
            abs_output = StarSet()
            for star in abs_input.stars:
                abs_output.stars = abs_output.stars.union(single_sigmoid_forward(star, self.approx_levels))

        return abs_output

    def backward(self, ref_state: RefinementState):
        """
        Update the RefinementState.

        Parameters
        ----------
        ref_state: RefinementState
            The RefinementState to update.
        """
        pass


class AbsNeuralNetwork(abc.ABC):
    """
    An abstract class used for our internal representation of a generic NeuralNetwork for Abstract Interpretation.
    It consists of a graph of AbsLayerNodes. The properties of the computational graph are specialized in the
    concrete classes. The method forward and backward calls the corresponding methods in the AbsLayerNodes following the
    correct order to compute the output AbsElement.

    Attributes
    ----------
    nodes : dict <str, LayerNode>
        Dictionary containing str keys and AbsLayerNodes values. It contains the nodes of the graph,
        the identifier of the node of interest is used as a key in the nodes dictionary.

    edges : dict <str, list <str>>
        Dictionary of identifiers of AbsLayerNodes, it contains for each nodes identified by the keys, the list of nodes
        connected to it.

    Methods
    ----------
    forward(AbsElement)
        Function which takes an AbsElement and compute the corresponding output AbsElement based on the AbsLayerNode
        of the network.

    backward(RefinementState)
        Function which takes a reference to the refinement state and update both it and the state of the AbsLayerNodes
        to control the refinement component of the abstraction.

    """

    def __init__(self):
        self.nodes = {}
        self.edges = {}

    def forward(self, abs_input: AbsElement) -> AbsElement:
        """
        Compute the output AbsElement based on the input AbsElement and the characteristics of the
        concrete abstract transformers.

        Parameters
        ----------
        abs_input : AbsElement
            The input abstract element.

        Returns
        ----------
        AbsElement
            The AbsElement resulting from the computation corresponding to the abstract transformer.
        """
        pass

    @abc.abstractmethod
    def backward(self, ref_state: RefinementState):
        """
        Update the RefinementState.

        Parameters
        ----------
        ref_state: RefinementState
            The RefinementState to update.
        """
        pass


class AbsSeqNetwork(AbsNeuralNetwork):
    """
    Concrete children of AbsNeuralNetwork representing a sequential AbsNeuralNetwork.
    It consists of a graph of LayerNodes. The computational graph of a SequentialNetwork must
    correspond to a standard list. The method forward and backward calls the corresponding methods
    in the AbsLayerNodes following the correct order to compute the output AbsElement.

    Attributes
    ----------
    identifier : str
        Identifier of the Sequential AbsNeuralNetwork.

    Methods
    -------
    add_node(LayerNode)
        Procedure to add a new AbsLayerNode to the sequential AbsNeuralNetwork.

    get_first_node()
        Procedure to extract the first AbsLayerNode of the sequential AbsNeuralNetwork.

    get_next_node(LayerNode)
        Procedure to get the next AbsLayerNode of the AbsNeuralNetwork given an input AbsLayerNode

    get_last_node()
        Procedure to extract the last AbsLayerNode of the sequential AbsNeuralNetwork.

    forward(AbsElement)
        Function which takes an AbsElement and compute the corresponding output AbsElement based on the AbsLayerNode
        of the network.

    backward(RefinementState)
        Function which takes a reference to the refinement state and update both it and the state of the AbsLayerNodes
        to control the refinement component of the abstraction.

    """

    def __init__(self, identifier: str):

        super().__init__()
        self.identifier = identifier

    def add_node(self, node: AbsLayerNode):
        """
        Procedure to add a new AbsLayerNode. In sequential network the new node must be connected directly to the
        previous node forming a list.

        Parameters
        ----------
        node : AbsLayerNode
            New node to add to the Sequential network.

        """

        if len(self.nodes.keys()) == 0:
            self.nodes[node.identifier] = node
            self.edges[node.identifier] = []
        else:
            previous_node_key = self.get_last_node().identifier
            self.nodes[node.identifier] = node
            self.edges[previous_node_key].append(node.identifier)
            self.edges[node.identifier] = []

    def get_first_node(self) -> AbsLayerNode:
        """
        Procedure to get the first AbsLayerNode of the network.

        Return
        ---------
        AbsLayerNode
            The first node of the network.

        """
        assert self.nodes

        keys = [key for key in self.nodes.keys()]
        for key in self.nodes.keys():
            for sub_key in self.nodes.keys():
                if sub_key in self.edges[key]:
                    keys.remove(sub_key)

        return self.nodes[keys[0]]

    def get_next_node(self, node: AbsLayerNode) -> AbsLayerNode:
        """
        Procedure to get the next AbsLayerNode of the network given an input AbsLayerNode.

        Return
        ---------
        LayerNode
            The next node of the network.

        """

        assert self.nodes

        next_node = None
        if node is not None:
            current_key = node.identifier
            if len(self.edges[current_key]) != 0:
                next_key = self.edges[current_key][0]
                next_node = self.nodes[next_key]
        else:
            next_node = self.get_first_node()

        return next_node

    def get_last_node(self) -> AbsLayerNode:
        """
        Procedure to get the last AbsLayerNode of the network.

        Return
        ---------
        AbsLayerNode
            The last node of the network.

        """

        assert self.nodes

        current_node = self.get_first_node()
        while self.get_next_node(current_node) is not None:
            current_node = self.get_next_node(current_node)

        return current_node

    def forward(self, abs_input: AbsElement) -> AbsElement:
        """
        Compute the output AbsElement based on the input AbsElement and the characteristics of the
        concrete abstract transformers.

        Parameters
        ----------
        abs_input : AbsElement
            The input abstract element.

        Returns
        ----------
        AbsElement
            The AbsElement resulting from the computation corresponding to the abstract transformer.
        """

        current_node = self.get_first_node()
        while current_node is not None:
            abs_input = current_node.forward(abs_input)
            current_node = self.get_next_node(current_node)

        return abs_input

    def backward(self, ref_state: RefinementState):
        """
        Update the RefinementState.

        Parameters
        ----------
        ref_state: RefinementState
            The RefinementState to update.
        """
        pass
